#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  Copyright (c) 2019.       Mike Herbert
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
"""
geoname database support routines.  Add locations to geoname DB, create geoname tables and indices.   
Provides a number of methods to lookup locations by name, feature, admin ID, etc.
"""
import logging
import os
import re
import sys
import time
from datetime import timedelta
from operator import itemgetter
from tkinter import messagebox

from geodata import GeoUtil, Loc, Country, MatchScore, Normalize, DB, QueryList
from geodata.GeoUtil import Query, Result, Entry, get_soundex


# from util import SpellCheck


class GeoDB:
    """
    geoname database routines.  Add items to geoname DB, look up items, create tables, indices
    """

    def __init__(self, db_path, spellcheck,
                 show_message: bool, exit_on_error: bool, set_speed_pragmas: bool, db_limit: int):
        """
            geoname data database init. Open database if present otherwise raise error
        # Args:
            db_path: full path to database file
            spellcheck: True if Spellcheck should be enabled.  NOT CURRENTLY SUPPORTED.  Must be False
            show_message: If True, show messagebox to user on error
            exit_on_error: If True, exit if significant error occurs
            set_speed_pragmas: If True, set DB pragmas for maximum performance. 
            db_limit: SQL LIMIT parameter
        # Raises:
            ValueError('Cannot open database'), ValueError('Database empty or corrupt')
        """
        self.logger = logging.getLogger(__name__)
        self.start = 0
        self.match = MatchScore.MatchScore()
        self.select_str = 'name, country, admin1_id, admin2_id, lat, lon, f_code, geoid, sdx'
        if spellcheck:
            raise ValueError('Spellcheck is not currently supported')

        self.spellcheck = spellcheck
        self.use_wildcards = True
        self.total_time = 0
        self.total_lookups = 0
        self.max_query_results = 50

        self.db_path = db_path
        # See if DB exists 
        if os.path.exists(db_path):
            db_existed = True
        else:
            db_existed = False

        self.db = DB.DB(db_filename=db_path, show_message=show_message, exit_on_error=exit_on_error)
        if self.db.err != '':
            self.logger.error(f"Error! cannot open database {db_path}.")
            raise ValueError('Cannot open database')

        # If DB was initially found 
        if db_existed:
            # Run sanity test on DB
            err = self.db.test_database('name', 'main.geodata', where='name = ? AND country = ?', args=('ba', 'fr'))

            if err:
                # DB failed sanity test
                self.logger.warning(f'DB error for {db_path}')

                if show_message:
                    if messagebox.askyesno('Error',
                                           f'Geoname database is empty or corrupt:\n\n {db_path} \n\nDo you want to delete it and rebuild?'):
                        messagebox.showinfo('', 'Deleting Geoname database')
                        self.db.conn.close()
                        os.remove(db_path)
                if exit_on_error:
                    sys.exit()
                else:
                    raise ValueError('Database empty or corrupt')

        if set_speed_pragmas:
            self.db.set_speed_pragmas()
        self.db_limit = db_limit
        self.db.order_string = ''
        self.db.limit_string = f'LIMIT {self.db_limit}'
        self.geoid_main_dict = {}  # Key is GEOID, Value is DB ID for entry
        self.geoid_admin_dict = {}  # Key is GEOID, Value is DB ID for entry
        self.place_type = ''

    def lookup_place(self, place: Loc) -> []:
        """
            **Lookup a place in geoname.org db**     
            Lookup is based on place.place_type as follows:  
                Loc.PlaceType.ADMIN1: does self.wide_search_admin1(place)  
                Loc.PlaceType.ADMIN2: does self.wide_search_admin2(place)  
                Loc.PlaceType.COUNTRY: does self.wide_search_country(place)  
                Loc.PlaceType.ADVANCED_SEARCH: does self.feature_search(place)  
                Otherwise: do self.wide_search_city(place)  
        # Args:   
            place: Loc instance.  Call Loc.parse_place() before calling lookup_place()   
  
        # Returns:   
            None.  
            Place.georow_list contains a list of matching entries.  
            Each entry has: Lat, Long, districtID (County or State or Province ID), and a match quality score  

        """
        self.start = time.time()
        place.result_type = Result.STRONG_MATCH

        if place.country_iso != '' and place.country_name == '':
            place.country_name = self.get_country_name(place.country_iso)

        target_feature = place.place_type

        # Lookup Place based on Place Type
        if place.place_type == Loc.PlaceType.ADMIN1:
            lookup_type = 'ADMIN1'
            self.wide_search_admin1(place)
        elif place.place_type == Loc.PlaceType.ADMIN2:
            lookup_type = 'ADMIN2'
            if place.admin1_id == '':
                self.wide_search_admin1_id(place=place)
            self.wide_search_admin2(place)
        elif place.place_type == Loc.PlaceType.COUNTRY:
            lookup_type = 'COUNTRY'
            self.wide_search_country(place)
        elif place.place_type == Loc.PlaceType.ADVANCED_SEARCH:
            self.feature_search(place)
            lookup_type = 'ADVANCED'
        else:
            # Lookup as City
            lookup_type = 'CITY'
            #if place.admin2_id == '':
            #    self.wide_search_admin2_id(place=place)
            self.wide_search_city(place)

        if place.georow_list:
            self.assign_scores(place, target_feature)
            self.logger.debug(f'LOOKUP: {len(place.georow_list)} matches for type={lookup_type}  '
                              f'targ={place.target} nm=[{place.get_five_part_title()}]\n')
        else:
            # self.logger.debug(f'LOOKUP. No match:for {lookup_type}  targ={place.target} nm=[{place.get_five_part_title()}]\n')
            place.georow_list = []
            
    def add_query_by_feature(self, query_list, lookup_target, iso):
        # Scan target to see if we can determine what feature type it is
        word, group = GeoUtil.get_feature_group(lookup_target)
        if word != '':
            targ = re.sub(word, '', lookup_target)
            query_list.append(Query(where="name LIKE ? AND country LIKE ? AND f_code LIKE ?",
                  args=(targ, iso, group),
                  result=Result.WORD_MATCH))
            self.logger.debug(f'Added Feature query name={targ} group={group}')

    def wide_search_city(self, place: Loc):
        """
        Search for city using place.target
        # Args:   
            place: Loc instance   
        # Returns:   
            None.  place.georow_list is updated with list of matches   
        """
        lookup_target = place.target
        if len(place.target) == 0:
            return

        pattern = self.create_wildcard(lookup_target)

        sdx = get_soundex(lookup_target)
        self.logger.debug(f'CITY lkp targ=[{lookup_target}] adm1 id=[{place.admin1_id}]'
                          f' adm2 id=[{place.admin2_id}] iso=[{place.country_iso}] patt =[{pattern}] sdx={sdx} pref={place.prefix}')

        query_list = []
        
        QueryList.QueryList.build_query_list(typ=QueryList.Typ.CITY, query_list=query_list, place=place)

        if len(place.country_iso) == 0:
            # NO COUNTRY - do lookup just by name 
            if lookup_target in pattern:
                query_list.append(Query(where="name = ?",
                                        args=(lookup_target,),
                                        result=Result.PARTIAL_MATCH))
            # lookup by wildcard name
            if '*' in lookup_target:
                query_list.clear()
                query_list.append(Query(where="name LIKE ?",
                                        args=(pattern,),
                                        result=Result.WILDCARD_MATCH))
            else:
                query_list.append(Query(where="name LIKE ?",
                                        args=(pattern,),
                                        result=Result.WORD_MATCH))

            # lookup by soundex
            query_list.append(Query(where="sdx = ?",
                                    args=(sdx,),
                                    result=Result.SOUNDEX_MATCH))
            
            place.georow_list, place.result_type = self.process_query_list(select_string=self.select_str,
                                                                           from_tbl='main.geodata',
                                                                           query_list=query_list)
            # self.logger.debug(place.georow_list)
            return

        if len(place.admin1_name) > 0:
            # lookup by name, ADMIN1, country
            if lookup_target in pattern:
                query_list.append(Query(
                    where="name = ? AND country = ? AND admin1_id = ?",
                    args=(lookup_target, place.country_iso, place.admin1_id),
                    result=Result.STRONG_MATCH))

            # lookup by wildcard name, ADMIN1, country
            if '*' in lookup_target:
                query_list.clear()

                query_list.append(Query(
                    where="name LIKE ? AND country = ? AND admin1_id = ?",
                    args=(pattern, place.country_iso, place.admin1_id),
                    result=Result.WILDCARD_MATCH))
            else:
                query_list.append(Query(
                    where="name LIKE ? AND country = ? AND admin1_id = ?",
                    args=(pattern, place.country_iso, place.admin1_id),
                    result=Result.WORD_MATCH))
        else:
            # No admin1 - lookup by name, country
            if '*' in lookup_target:
                query_list.clear()

                query_list.append(Query(where="name LIKE ? AND country = ?",
                                        args=(pattern, place.country_iso),
                                        result=Result.WILDCARD_MATCH))
            else:
                query_list.append(Query(where="name LIKE ? AND country = ?",
                                        args=(pattern, place.country_iso),
                                        result=Result.WORD_MATCH))

        # lookup by Soundex , country
        #query_list.append(Query(where="sdx = ? AND country = ?",
        #                        args=(sdx, place.country_iso),
        #                        result=Result.SOUNDEX_MATCH))
        
        # Try to find feature type and lookup by that
        self.add_query_by_feature(query_list, lookup_target, place.country_iso)

        # Try each query in list
        place.georow_list, place.result_type = self.process_query_list(select_string=self.select_str, from_tbl='main.geodata',
                                                                       query_list=query_list)

    def wide_search_admin2(self, place: Loc):
        """
        Search for Admin2 using place.admin2_name

        # Args:   
            place:   

        # Returns:
            None.  place.georow_list is updated with list of matches   
        """
        save_target = place.target

        place.target = place.admin2_name
        if len(place.target) == 0:
            return

        query_list = []
        QueryList.QueryList.build_query_list(typ=QueryList.Typ.ADMIN2, query_list=query_list, place=place)

        # self.logger.debug(f'Admin2 lookup=[{lookup_target}] country=[{place.country_iso}]')
        place.georow_list, place.result_type = self.process_query_list(select_string=self.select_str, from_tbl='main.admin', query_list=query_list)

        if len(place.georow_list) == 0:
            # Try city rather than County match.
            save_admin2 = place.admin2_name
            place.city1 = place.admin2_name
            place.admin2_name = ''
            # self.logger.debug(f'Try admin2 as city: [{place.target}]')

            self.wide_search_city(place)

            if len(place.georow_list) == 0:
                #  not found.  restore admin
                place.admin2_name = save_admin2
                place.city1 = ''
            else:
                # Found match as a City
                place.place_type = Loc.PlaceType.CITY
                match_adm1 = self.get_admin1_name_direct(admin1_id=place.georow_list[0][Entry.ADM1], iso=place.country_iso)
                # self.logger.debug(f'pl_iso [{place.country_iso}] pl_adm1 {place.admin1_name} match_adm1=[{match_adm1}] ')
                if place.admin1_name != match_adm1 and '*' not in place.admin1_name.title():
                    # place.prefix = place.admin1_name.title()
                    place.admin1_name = ''
                return

        place.target = save_target

    def wide_search_admin1(self, place: Loc):
        """
        Search for Admin1 using place.admin1_name

        # Args:   
            place:   

        # Returns:
            None.  place.georow_list is updated with list of matches   
        """
        save_target = place.target
        place.target = place.admin1_name
        if len(place.target) == 0:
            return

        query_list = []
        QueryList.QueryList.build_query_list(typ=QueryList.Typ.ADMIN1, query_list=query_list, place=place)
        place.georow_list, place.result_type = self.process_query_list(select_string=self.select_str, from_tbl='main.admin', query_list=query_list)

        # Sort places in match_score order
        if len(place.georow_list) > 0:
            self.assign_scores(place, 'ADM1')
            sorted_list = sorted(place.georow_list, key=itemgetter(GeoUtil.Entry.SCORE))
            place.admin1_id = sorted_list[0][Entry.ADM1]
            # self.logger.debug(f'Found adm1 id = {place.admin1_id}')

        place.target = save_target

    def wide_search_country(self, place: Loc):
        """
        Search for Country using country_iso

        # Args:   
            place:   

        # Returns:
            None.  place.georow_list is updated with list of matches   
        """
        lookup_target = place.country_iso
        if len(lookup_target) == 0:
            return
        sdx = get_soundex(lookup_target)

        # Try each query until we find a match - each query gets less exact
        query_list = [
            Query(where="country = ? AND f_code = ? ",
                  args=(place.country_iso, 'ADM0'),
                  result=Result.STRONG_MATCH),
            Query(where="sdx = ?  AND f_code=?",
                  args=(sdx, 'ADM0'),
                  result=Result.SOUNDEX_MATCH)
            ]

        place.georow_list, place.result_type = self.process_query_list(select_string=self.select_str, from_tbl='main.admin', query_list=query_list)

    def wide_search_admin1_id(self, place: Loc):
        """
        Search for Admin1 ID using place.admin1_name

        # Args:   
            place:   

        # Returns:
            None.  place.admin1_id and place.country_iso are updated with best match 
        """
        lookup_target = place.admin1_name
        pattern = self.create_admin1_wildcard(lookup_target)
        if len(lookup_target) == 0:
            return
        save_prefix = place.prefix

        query_list = []

        # Try each query then calculate best match - each query gets less exact
        if place.country_iso == '':
            query_list.append(Query(where="name = ?  AND f_code = ? ",
                                    args=(lookup_target, 'ADM1'),
                                    result=Result.STRONG_MATCH))

            query_list.append(Query(where="name LIKE ? AND f_code = ?",
                                    args=(pattern, 'ADM1'),
                                    result=Result.WILDCARD_MATCH))
        else:
            query_list.append(Query(where="name = ? AND country = ? AND f_code = ? ",
                                    args=(lookup_target, place.country_iso, 'ADM1'),
                                    result=Result.STRONG_MATCH))
            query_list.append(Query(where="name LIKE ? AND country = ?  AND f_code = ?",
                                    args=(pattern, place.country_iso, 'ADM1'),
                                    result=Result.WILDCARD_MATCH))

        query_list.append(Query(where="name = ?  AND f_code = ?",
                                args=(lookup_target, 'ADM1'),
                                result=Result.SOUNDEX_MATCH))

        place.georow_list, place.result_type = self.process_query_list(select_string=self.select_str, from_tbl='main.admin', query_list=query_list)

        self.assign_scores(place, 'ADM1')
        # Sort places in match_score order
        if len(place.georow_list) > 0:
            sorted_list = sorted(place.georow_list, key=itemgetter(GeoUtil.Entry.SCORE))
            score = sorted_list[0][Entry.SCORE]
            place.admin1_id = sorted_list[0][Entry.ADM1]

            self.logger.debug(f'Found adm1 id = {place.admin1_id}  score={score:.1f}')
            # Fill in Country ISO
            if place.country_iso == '':
                place.country_iso = sorted_list[0][Entry.ISO]
        place.prefix = save_prefix

    def wide_search_admin2_id(self, place: Loc):
        """
             Search for Admin2 ID using place.admin2_name

        # Args:   
            place:   

        # Returns:
            None.  place.admin2_id is updated with best match 
        """
        lookup_target = place.admin2_name
        save_prefix = place.prefix
        # pattern = self.create_county_wildcard(lookup_target)
        if len(lookup_target) == 0:
            return

        # Try each query until we find a match - each query gets less exact
        query_list = []
        if len(place.admin1_id) > 0:
            query_list.append(Query(where="name = ? AND country = ? AND admin1_id=? AND f_code=?",
                                    args=(lookup_target, place.country_iso, place.admin1_id, 'ADM2'),
                                    result=Result.STRONG_MATCH))
            query_list.append(Query(where="name LIKE ? AND country = ? and admin1_id = ? AND f_code=?",
                                    args=(lookup_target, place.country_iso, place.admin1_id, 'ADM2'),
                                    result=Result.WILDCARD_MATCH))
            query_list.append(Query(where="name LIKE ? AND country = ? and admin1_id = ? AND f_code=?",
                                    args=(self.create_county_wildcard(lookup_target), place.country_iso, place.admin1_id, 'ADM2'),
                                    result=Result.WILDCARD_MATCH))
        else:
            query_list.append(Query(where="name = ? AND country = ? AND f_code=?",
                                    args=(lookup_target, place.country_iso, 'ADM2'),
                                    result=Result.STRONG_MATCH))
            query_list.append(Query(where="name LIKE ? AND country = ? AND f_code=?",
                                    args=(lookup_target, place.country_iso, 'ADM2'),
                                    result=Result.WILDCARD_MATCH))
            query_list.append(Query(where="name LIKE ? AND country = ? AND f_code=?",
                                    args=(self.create_county_wildcard(lookup_target), place.country_iso, 'ADM2'),
                                    result=Result.WILDCARD_MATCH))

        place.georow_list, place.result_type = self.process_query_list(select_string=self.select_str, from_tbl='main.admin', query_list=query_list)

        self.assign_scores(place, 'ADM2')

        if place.result_type == Result.STRONG_MATCH:
            row = place.georow_list[0]
            place.admin2_id = row[Entry.ADM2]
        place.prefix = save_prefix

    def get_admin1_alt_name(self, place: Loc) -> (str, str):
        """
             Get Admin1 name from alternate name table 

        # Args:   
            place:   place instance.  place.admin1_id is used for lookup

        # Returns:
            None.  place.admin2_id is updated with best match 
        """

        lookup_target = place.admin1_id
        if len(lookup_target) == 0:
            return '', ''

        # Try each query until we find a match - each query gets less exact
        query_list = [
            Query(where="admin1_id = ? AND country = ?  AND f_code = ? ",
                  args=(lookup_target, place.country_iso, 'ADM1'),
                  result=Result.STRONG_MATCH)
            ]
        row_list, res = self.process_query_list(select_string=self.select_str, from_tbl='main.admin', query_list=query_list)

        if len(row_list) > 0:
            row = row_list[0]
            admin1_name, lang = self.get_alt_name(row[Entry.ID])
            return admin1_name, lang
        else:
            return '', ''

    def get_admin1_name_direct(self, admin1_id, iso) -> str:
        """
        Search for Admin1 name using admin1_id (rather than place instance)

        # Args:   
            admin1_id: Admin1 ID   
            iso: country ISO     

        # Returns:
            None.  place.admin1_id and place.country_iso are updated with best match 
        """
        if len(admin1_id) == 0:
            return ''

        # Try each query until we find a match - each query gets less exact
        query_list = [
            Query(where="admin1_id = ? AND country = ?  AND f_code = ? ",
                  args=(admin1_id, iso, 'ADM1'),
                  result=Result.STRONG_MATCH)
            ]
        row_list, res = self.process_query_list(select_string=self.select_str, from_tbl='main.admin', query_list=query_list)

        if len(row_list) > 0:
            row = row_list[0]
            return row[Entry.NAME]
        else:
            return ''

    def get_admin2_name_direct(self, admin1_id, admin2_id, iso) -> str:
        """
        Search for Admin2 name using admin2_id and admin1_id (rather than place instance)

        # Args:   
            admin1_id: Admin1 ID   
            admin2_id: Admin2 ID   
            iso: country ISO     

        # Returns:
            None.  place.admin1_id and place.country_iso are updated with best match 
        """
        if len(admin2_id) == 0:
            return ''

        # Try each query until we find a match - each query gets less exact
        query_list = [
            Query(where="admin2_id = ? AND country = ? AND admin1_id = ?",
                  args=(admin2_id, iso, admin1_id),
                  result=Result.STRONG_MATCH),
            Query(where="admin2_id = ? AND country = ?",
                  args=(admin2_id, iso),
                  result=Result.PARTIAL_MATCH)]

        row_list, res = self.process_query_list(select_string=self.select_str, from_tbl='main.admin', query_list=query_list)

        if len(row_list) > 0:
            row = row_list[0]
            return row[Entry.NAME]
        else:
            return ''

    def get_geoid(self, place: Loc) -> None:
        """
             Search for location using Geoid in place.target

        # Args:   
            place:   place instance.  Uses  Geoid in place.target for lookup

        # Returns:
            None.   Place instance is updated with DB results
        """
        result_place: Loc = Loc.Loc()

        query_list = [
            Query(where="geoid = ? ",
                  args=(place.target,),
                  result=Result.STRONG_MATCH)
            ]
        place.georow_list, place.result_type = self.process_query_list(select_string=self.select_str,
                                                                       from_tbl='main.geodata', query_list=query_list)
        if len(place.georow_list) == 0:
            place.georow_list, place.result_type = self.process_query_list(select_string=self.select_str,
                                                                           from_tbl='main.admin', query_list=query_list)
        else:
            place.georow_list = place.georow_list[:1]
            place.result_type = GeoUtil.Result.STRONG_MATCH

        # Add search quality score to each entry
        for idx, rw in enumerate(place.georow_list):
            self.copy_georow_to_place(row=rw, place=result_place)
            update = list(rw)
            update.append(1)  # Extend list row and assign score
            result_place.prefix = ''
            result_name = result_place.get_long_name(None)
            score = 0.0

            # Remove any words from Prefix that are in result name
            place.prefix = Loc.Loc.prefix_cleanup(place.prefix, result_name)

            update[GeoUtil.Entry.SCORE] = int(score * 100)
            place.georow_list[idx] = tuple(update)

    def get_admin1_name(self, place: Loc) -> str:
        """
             Get admin1 name using place.admin1_id

        # Args:   
            place:   place instance.  Uses  place.admin1_id for lookup

        # Returns:
            Admin1 name.  Place instance admin1_name is updated with DB result
        """
        place.admin1_name = self.get_admin1_name_direct(place.admin1_id, place.country_iso)
        return place.admin1_name

    def get_admin2_name(self, place: Loc) -> str:
        """
             Get admin2 name using place.admin1_id and place.admin2_id

        # Args:   
            place:   place instance.  Uses  place.admin1_id and place.admin2_id for lookup

        # Returns:
            Admin2 name.  Place instance admin2_name is updated with DB result
        """
        place.admin2_name = self.get_admin2_name_direct(place.admin1_id, place.admin2_id, place.country_iso)
        return place.admin2_name

    def lookup_main_dbid(self, place: Loc) -> None:
        """Search for DB ID in main table"""
        query_list = [
            Query(where="id = ? ",
                  args=(place.target,),
                  result=Result.STRONG_MATCH)
            ]
        place.georow_list, place.result_type = self.process_query_list(select_string=self.select_str,
                                                                       from_tbl='main.geodata', query_list=query_list)

    def lookup_admin_dbid(self, place: Loc) -> None:
        """Search for DB ID in admin table"""
        query_list = [
            Query(where="id = ? ",
                  args=(place.target,),
                  result=Result.STRONG_MATCH)
            ]
        place.georow_list, place.result_type = self.process_query_list(select_string=self.select_str, from_tbl='main.admin', query_list=query_list)

    def get_country_name(self, iso: str) -> str:
        """
             return country name for specified ISO code 

        # Args:   
            iso:   Country ISO code

        # Returns:
            Country name or ''
        """
        if len(iso) == 0:
            return ''

        # Try each query until we find a match - each query gets less exact
        query_list = [
            Query(where="country = ? AND f_code = ? ",
                  args=(iso, 'ADM0'),
                  result=Result.STRONG_MATCH)]

        row_list, res = self.process_query_list(select_string=self.select_str, from_tbl='main.admin', query_list=query_list)

        if len(row_list) > 0:
            res = row_list[0][Entry.NAME]
            if iso == 'us':
                res = 'United States'
        else:
            res = ''
        return res

    def get_country_iso(self, place: Loc) -> str:
        """
             return country ISO code for place.country_name   

        # Args:   
            place:   place instance.  looks up by place.country_name   

        # Returns:   
            Country ISO or ''.  If found, update place.country_name with DB country name   
        """
        lookup_target, modified = Normalize.country_normalize(place.country_name)
        if len(lookup_target) == 0:
            return ''
        query_list = []

        # Add queries - each query gets less exact
        query_list.append(Query(where="name = ? AND f_code = ? ",
                                args=(lookup_target, 'ADM0'),
                                result=Result.STRONG_MATCH))

        if self.spellcheck:
            pattern = self.spellcheck.fix_spelling(lookup_target)
            query_list.append(Query(where="name LIKE ?  AND f_code = ? ",
                                    args=(pattern, 'ADM0'),
                                    result=Result.WILDCARD_MATCH))

        place.georow_list, place.result_type = self.process_query_list(select_string=self.select_str, from_tbl='main.admin', query_list=query_list)

        self.assign_scores(place, 'ADM0')

        if place.result_type == Result.STRONG_MATCH:
            iso = place.georow_list[0][Entry.ISO]
            place.country_name = place.georow_list[0][Entry.NAME]
        else:
            iso = ''

        return iso

    def feature_search(self, place: Loc):
        """
                Feature search - lookup by name, ISO Country and Feature class   
                e.g. place.target='d*'   
                    place.country_iso='gb'   
                    place.feature='CSTL'   
        # Args:   
            place: Uses place.target as lookup target, place.feature as feature target,   
                place.country_iso as country code target.   

        # Returns: 
            None.  place.georow_list has list of matching georows   

        """

        lookup_target = place.target
        if len(lookup_target) == 0:
            return
        pattern = self.create_wildcard(lookup_target)
        feature_pattern = self.create_wildcard(place.feature)
        self.logger.debug(f'Advanced Search. Targ=[{pattern}] feature=[{feature_pattern}]'
                          f'  iso=[{place.country_iso}] ')

        if len(place.feature) > 0:
            query_list = [
                Query(where="name LIKE ? AND country LIKE ? AND f_code LIKE ?",
                      args=(pattern, place.country_iso, feature_pattern),
                      result=Result.PARTIAL_MATCH)]
        else:
            query_list = [
                Query(where="name LIKE ? AND country LIKE ?",
                      args=(pattern, place.country_iso),
                      result=Result.PARTIAL_MATCH)]

        # Search main DB
        place.georow_list, place.result_type = self.process_query_list(select_string=self.select_str,
                                                                       from_tbl='main.geodata',
                                                                       query_list=query_list)

        # self.logger.debug(f'main Result {place.georow_list}')

        # Search admin DB
        admin_list, place.result_type = self.process_query_list(select_string=self.select_str, from_tbl='main.admin', query_list=query_list)
        place.georow_list.extend(admin_list)
        # self.logger.debug(f'admin Result {place.georow_list}')

    def copy_georow_to_place(self, row, place: Loc):
        """
        Copy data from DB row into place instance   
        Country, admin1_id, admin2_id, city, lat/lon, feature, geoid are updated if available   
        #Args:   
            row: georow from geoname database   
            place: Loc instance   

        #Returns:   
            None.  Place instance is updated with data from georow   
        """
        place.admin1_id = ''
        place.admin2_id = ''
        place.city1 = ''

        place.country_iso = str(row[Entry.ISO])
        place.country_name = str(self.get_country_name(row[Entry.ISO]))
        place.lat = row[Entry.LAT]
        place.lon = row[Entry.LON]
        place.feature = str(row[Entry.FEAT])
        place.geoid = str(row[Entry.ID])
        place.prefix = row[Entry.PREFIX]
        place.place_type = Loc.PlaceType.CITY

        if place.feature == 'ADM0':
            place.place_type = Loc.PlaceType.COUNTRY
            pass
        elif place.feature == 'ADM1':
            place.admin1_id = row[Entry.ADM1]
            place.place_type = Loc.PlaceType.ADMIN1
        elif place.feature == 'ADM2':
            place.admin1_id = row[Entry.ADM1]
            place.admin2_id = row[Entry.ADM2]
            place.place_type = Loc.PlaceType.ADMIN2
        else:
            place.admin1_id = row[Entry.ADM1]
            place.admin2_id = row[Entry.ADM2]
            place.city1 = row[Entry.NAME]

        place.admin1_name = str(self.get_admin1_name(place))
        place.admin2_name = str(self.get_admin2_name(place))
        if place.admin2_name is None:
            place.admin2_name = ''
        if place.admin1_name is None:
            place.admin1_name = ''

        place.city1 = str(place.city1)
        if place.city1 is None:
            place.city1 = ''

        try:
            place.score = row[Entry.SCORE]
        except IndexError as e:
            pass

    def clear_geoname_data(self):
        """
        Delete geodata table and admin table from database
        """
        for tbl in ['geodata', 'admin']:
            # noinspection SqlWithoutWhere
            self.db.delete_table(tbl)

    def create_geoid_index(self):
        """
        Create database indices for GEOID
        """
        self.db.create_index(create_index_sql='CREATE INDEX IF NOT EXISTS geoid_idx ON geodata(geoid)')
        self.db.create_index(create_index_sql='CREATE INDEX IF NOT EXISTS admgeoid_idx ON admin(geoid)')
        self.db.create_index(create_index_sql='CREATE INDEX IF NOT EXISTS altnamegeoid_idx ON altname(geoid)')

    def create_indices(self):
        """
        Create indices for geoname database
        """
        self.db.create_index(create_index_sql='CREATE INDEX IF NOT EXISTS name_idx ON geodata(name, country )')
        self.db.create_index(create_index_sql='CREATE INDEX IF NOT EXISTS admin1_idx ON geodata(admin1_id )')
        self.db.create_index(create_index_sql='CREATE INDEX IF NOT EXISTS sdx_idx ON geodata(sdx )')

        self.db.create_index(create_index_sql='CREATE INDEX IF NOT EXISTS adm_name_idx ON admin(name, country )')
        self.db.create_index(create_index_sql='CREATE INDEX IF NOT EXISTS adm_admin1_idx ON admin(admin1_id, f_code)')
        self.db.create_index(create_index_sql='CREATE INDEX IF NOT EXISTS adm_admin2_idx ON admin(admin1_id, admin2_id)')
        self.db.create_index(create_index_sql='CREATE INDEX IF NOT EXISTS adm_country_idx ON admin(country, f_code)')
        self.db.create_index(create_index_sql='CREATE INDEX IF NOT EXISTS adm_sdx_idx ON admin(sdx )')

    @staticmethod
    def make_georow(name: str, iso: str, adm1: str, adm2: str, lat: float, lon: float, feat: str, geoid: str, sdx: str) -> ():
        """
        Create a georow based on arguments
        # Args:
            name:   
            iso:   
            adm1: admin1 id   
            adm2: admin2_id   
            lat:   
            lon: 
            feat:     
            geoid:   
            sdx:   

        # Returns:    
            georow

        """
        res = (name, iso, adm1, adm2, lat, lon, feat, geoid, sdx)
        return res

    def get_row_count(self) -> int:
        """
        Get row count of main.geodata
        :return:
        """
        return self.db.get_row_count('main.geodata')

    @staticmethod
    def create_wildcard(pattern):
        """
        Create wildcard pattern.  Convert * to %.  Add % on end   
        #Args:   
            pattern:   

        #Returns: wildcard pattern   

        """
        if '*' in pattern:
            return re.sub(r"\*", "%", pattern)
        else:
            return f'%{pattern}%'

    @staticmethod
    def create_admin1_wildcard(pattern):
        """
        Create wildcard pattern.  Convert * to %.  Add % on end.  Currently this  
        is the same at create_wildcard   
        #Args:   
            pattern:    

        #Returns: wildcard pattern   

        """
        if '*' in pattern:
            return re.sub(r"\*", "%", pattern)
        else:
            return f'{pattern}%'

    @staticmethod
    def create_county_wildcard(pattern):
        """
        create wildcard for county lookup  
        #Args:   
            pattern:    

        #Returns:  

        """
        # Try pattern with 'shire' removed
        pattern = re.sub(r"shire", "", pattern)
        # Create SQL wildcard pattern (convert * to %).  Add % on end
        if '*' in pattern:
            return re.sub(r"\*", "%", pattern)
        else:
            return f'{pattern}'

    def close(self):
        """
        Close database.  Set optimize pragma
        """
        self.db.set_optimize_pragma()
        self.logger.info(f'TOTAL DATABASE LOOKUP TIME: {str(timedelta(seconds=self.total_time))}')
        if self.total_time > 0:
            self.logger.info(f'Lookups per second {self.total_lookups / self.total_time:.0f}')
        self.logger.info('Closing Database')
        self.db.conn.close()

    def set_display_names(self, temp_place):
        """
            See if there is an alternate name entry for this place   

        #Args:   
            temp_place: place instance   

        #Returns: None   

        """
        place_lang = Country.Country.get_lang(temp_place.country_iso)
        res, lang = self.get_alt_name(temp_place.geoid)
        if res != '' and (lang == place_lang or lang == 'ut8'):
            temp_place.city1 = res

        res, lang = self.get_admin1_alt_name(temp_place)
        if res != '' and (lang == place_lang or lang == 'ut8'):
            temp_place.admin1_name = res

    def get_alt_name(self, geoid) -> (str, str):
        """
        Retrieve alternate names for specified GEOID   

        #Args:    
            geoid: Geoid to get alternate names for   

        #Returns: 
            row_list from DB matches   

        """
        query_list = [
            Query(where="geoid = ?",
                  args=(geoid,),
                  result=Result.STRONG_MATCH)]
        select = 'name, lang'
        row_list, res = self.process_query_list(select_string=select, from_tbl='main.altname', query_list=query_list)
        if len(row_list) > 0:
            return row_list[0][0], row_list[0][1]
        else:
            return '', ''

    def insert(self, geo_row: (), feat_code: str):
        """
        Insert a geo_row into geonames database   
        #Args:   
            geo_row: row to insert   
            feat_code: Geonames feature code of item   
        #Returns:   
            row_id for inserted row   
        """
        # We split the data into 2  tables, 1) Admin: ADM0/ADM1/ADM2,  and 2) city data
        if feat_code == 'ADM1' or feat_code == 'ADM0' or feat_code == 'ADM2':
            sql = ''' INSERT OR IGNORE INTO admin(name,country, admin1_id,admin2_id,lat,lon,f_code, geoid, sdx)
                      VALUES(?,?,?,?,?,?,?,?,?) '''
            row_id = self.db.execute(sql, geo_row)
            # Add name to dictionary.  Used by AlternateNames for fast lookup during DB build
            self.geoid_admin_dict[geo_row[Entry.ID]] = row_id
        else:
            sql = ''' INSERT OR IGNORE INTO geodata(name, country, admin1_id, admin2_id, lat, lon, f_code, geoid, sdx)
                      VALUES(?,?,?,?,?,?,?,?,?) '''
            row_id = self.db.execute(sql, geo_row)
            # Add name to dictionary.  Used by AlternateNames for fast lookup during DB build
            self.geoid_main_dict[geo_row[Entry.ID]] = row_id

        # Add item to Spell Check dictionary
        if self.spellcheck:
            self.spellcheck.insert(geo_row[Entry.NAME], geo_row[Entry.ISO])

        return row_id

    def insert_alternate_name(self, alternate_name: str, geoid: str, lang: str):
        """
        Add alternate name to altname table
        #Args:   
            alternate_name: alternate name to add for this geoid   
            geoid: geonames.org geoid   
            lang: ISO lang code for this entry   

        #Returns: None   

        """
        row = (alternate_name, lang, geoid)
        sql = ''' INSERT OR IGNORE INTO altname(name,lang, geoid)
                  VALUES(?,?,?) '''
        self.db.execute(sql, row)

    def insert_version(self, db_version: int):
        """
        Insert DB version into Database.  This is used to track when DB schema changes   
        #Args:   
            db_version: Version of this DB schema   

        #Returns: None   

        """
        self.db.begin()

        sql = ''' INSERT OR IGNORE INTO version(version)
                  VALUES(?) '''
        args = (db_version,)
        self.db.execute(sql, args)
        self.db.commit()

    def get_db_version(self) -> int:
        """
        Get schema version of database   
        #Returns: 
            schema version of database   

        """
        # If version table does not exist, this is V1
        if self.db.table_exists('version'):
            # query version ID
            query_list = [
                Query(where="version like ?",
                      args=('%',),
                      result=Result.STRONG_MATCH)]
            select_str = '*'
            row_list, res = self.process_query_list(select_string=select_str, from_tbl='main.version', query_list=query_list)
            if len(row_list) > 0:
                ver = int(row_list[0][1])
                self.logger.debug(f'Database Version = {ver}')
                return ver

        # No version table, so this is V1
        self.logger.debug('No version table.  Version is 1')
        return 1

    def process_query_list(self, select_string, from_tbl: str, query_list: [Query]):
        """
        Do a lookup for each query in the query_list.  Stop when self.max_query_results is reached    
        #Args:   
            select_string: SQL select string   
            from_tbl: Table to query   
            query_list: A list of SQL queries   

        #Returns: 
            tuple of row_list and result_type   

        """
        # Perform each SQL query in the list
        row_list = []

        result_type = Result.NO_MATCH
        for query in query_list:
            # Skip query if it's a wildcard and wildcards are disabled
            if self.use_wildcards is False and (query.result == Result.WILDCARD_MATCH or query.result == Result.SOUNDEX_MATCH):
                continue
            start = time.time()

            if query.result == Result.WORD_MATCH:
                result_list2 = self.word_match(select_string, query.where, from_tbl,
                                               query.args)
                if row_list:
                    row_list.extend(result_list2)
                else:
                    row_list = result_list2
            else:
                result_list = self.db.select(select_string, query.where, from_tbl,
                                             query.args)

                if row_list:
                    row_list.extend(result_list)
                else:
                    row_list = result_list

            if len(row_list) > 0:
                result_type = query.result

            elapsed = time.time() - start
            self.total_time += elapsed
            self.total_lookups += 1
            if elapsed > .005:
                if query.result == Result.WORD_MATCH:
                    self.logger.debug(f'Slow word lookup. Time={elapsed:.4f}  '
                                      f'len {len(row_list)} from {from_tbl} '
                                      f'where {query.where} val={query.args} ')
                else:
                    self.logger.debug(f'Slow lookup. Time={elapsed:.4f}  '
                                      f'len {len(row_list)} from {from_tbl} '
                                      f'where {query.where} val={query.args} ')
                    
            if len(row_list) > self.max_query_results:
                break

        return row_list, result_type

    def word_match(self, select_string, where, from_tbl, args):
        """
        Perform a wildcard match on each word in args[0], and then   
        merges the results into a single result.  During the merge, we note if   
        a duplicate occurs, and mark that as a higher priority result.  We  
        also note if an individual word has too many results, as we will drop   
        those results from the final list after doing the priority checks.   
        This should kill off common words from the results, while still   
        preserving combinations.   

        For example, searching for "Village of Bay", will find all three words   
        to have very many results, but the combinations of "Village" and "Bay"   
        or "Village" and "of" or "Bay" and "of" will show up in the results.   

        The order of the words will also not matter, so results should contain   
        "City of Bay Village", "Bay Village" etc.   
        
        #Args:   
            select_string:    
            where:  
            from_tbl:  
            args: args[0] has the words to search for   

        #Returns:   

        """
        words = args[0].split()
        results = []  # the entire merged list of result rows
        res_flags = []  # list of flags, matching results list, 'True' to keep
        if len(words) == 0:
            return []
        
        for word in words:
            # redo tuple for each word; select_string still has LIKE
            n_args = (f'%{word.strip()}%', *args[1:])
            result = self.db.select(select_string, where, from_tbl, n_args)
            for idx, row in enumerate(result):
                if idx > 50:
                    break
                # check if already in overall list
                for indx, r_row in enumerate(results):
                    if row[Entry.ID] == r_row[Entry.ID]:
                        # if has same ID as in overall list, mark to keep
                        res_flags[indx] = True
                        break
                else:  # this result row did not match anything
                    results.append(row)  # add it to overall list
                    # if reasonable number of results for this word, flag to
                    # keep the result
                    res_flags.append(len(result) < 50)
        # strip out any results not flagged (too many to be interesting)
        result = [results[indx] for indx in range(len(results)) if
                  res_flags[indx]]
        return result

    def assign_scores(self, place, target_feature):
        """
            Assign match score to each result in list   
        # Args:   
            place:   
            target_feature:  The feature type we were searching for   
        """
        result_place: Loc = Loc.Loc()

        min_score = 9999
        original_prefix = place.prefix #+ ' ' + place.extra   + ' ' + place.target

        # Remove redundant terms in prefix by converting it to dictionary (then back to list)
        # prefix_list = list(dict.fromkeys(original_prefix.split(' ')))
        # original_prefix = ' '.join(list(prefix_list))

        # Add search quality score and prefix to each entry
        for idx, rw in enumerate(place.georow_list):
            place.prefix = original_prefix
            self.copy_georow_to_place(row=rw, place=result_place)
            result_place.original_entry = result_place.get_long_name(None)

            if len(place.prefix) > 0 and result_place.prefix == '':
                result_place.prefix = ' '
                result_place.prefix_commas = ','
            else:
                result_place.prefix = ''

            # Remove items in prefix that are in result
            if place.place_type != Loc.PlaceType.ADVANCED_SEARCH:
                result_name = result_place.get_long_name(None)
                place.prefix = Loc.Loc.prefix_cleanup(place.prefix, result_name)
            else:
                place.updated_entry = place.get_long_name(None)

            score = self.match.match_score(target_place=place, result_place=result_place)

            if result_place.feature == target_feature:
                score -= 10

            min_score = min(min_score, score)

            # Convert row tuple to list and extend so we can assign score
            update = list(rw)
            if len(update) < GeoUtil.Entry.SCORE + 1:
                update.append(1)
            update[GeoUtil.Entry.SCORE] = score

            result_place.prefix = Normalize.normalize(place.prefix, True)
            update[GeoUtil.Entry.PREFIX] = result_place.prefix
            place.georow_list[idx] = tuple(update)  # Convert back from list to tuple

        if min_score < MatchScore.Score.VERY_GOOD + 2:
            place.result_type = GeoUtil.Result.STRONG_MATCH
            
        if min_score > MatchScore.Score.POOR:
            place.georow_list.clear()

    def create_tables(self):
        """
        Create all the tables needed for the geoname database
        """
        # name, country, admin1_id, admin2_id, lat, lon, f_code, geoid
        sql_geodata_table = """CREATE TABLE IF NOT EXISTS geodata    (
                id           integer primary key autoincrement not null,
                name     text,
                country     text,
                admin1_id     text,
                admin2_id text,
                lat      text,
                lon       text,
                f_code      text,
                geoid      text,
                sdx     text
                                    );"""

        # name, country, admin1_id, admin2_id, lat, lon, f_code, geoid
        sql_admin_table = """CREATE TABLE IF NOT EXISTS admin    (
                id           integer primary key autoincrement not null,
                name     text,
                country     text,
                admin1_id     text,
                admin2_id text,
                lat      text,
                lon       text,
                f_code      text,
                geoid      text,
                sdx     text
                                    );"""

        # name, country, admin1_id, admin2_id, lat, lon, f_code, geoid
        sql_alt_name_table = """CREATE TABLE IF NOT EXISTS altname    (
                id           integer primary key autoincrement not null,
                name     text,
                lang     text,
                geoid      text
                                    );"""

        # name, country, admin1_id, admin2_id, lat, lon, f_code, geoid
        sql_version_table = """CREATE TABLE IF NOT EXISTS version    (
                id           integer primary key autoincrement not null,
                version     integer
                                    );"""

        for tbl in [sql_geodata_table, sql_admin_table, sql_version_table, sql_alt_name_table]:
            self.db.create_table(tbl)



# -*- coding: utf-8 -*-
#  Copyright (c) 2019.       Mike Herbert
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
""" Build database for geonames.org data """
import csv
import logging
import os
import sys
import time
from collections import namedtuple
from tkinter import messagebox
from typing import Dict

# import SpellCheck
from geodata import GeoUtil, Loc, Country, GeoDB, Normalize, CachedDictionary, AlternateNames

DB_MINIMUM_RECORDS = 1000


class GeodataBuild:
    """
    Read in geonames.org geo data files, filter them and place the entries in a sqlite db.

    The files must be downloaded from geonames.org and used according to their usage rules.
    
    geoname files: allCountries.txt,  alternateNamesV2.txt

    The geoname file is filtered to only include the countries and feature codes specified in: 1) the country_list,
      2) the feature_list, and 3) the language list.  The import also creates new Feature codes: PP1M, P1HK, and P10K using   
     geoname population data (Pop > 1M, Pop > 100K, Pop > 10K).   

    """

    def __init__(self, directory: str, progress_bar, enable_spell_checker,
                 show_message, exit_on_error, languages_list_dct, feature_code_list_dct, supported_countries_dct):
        """
        Read in datafiles needed for geodata, filter them and create a sql db.
        Filter dictionary examples:   
            languages_list_dct={'fr','de'}
            feature_code_list_dct={'PPL', 'ADM1', 'CSTL'}
            supported_countries_dct = {'us','gb','at'}
        # Args:
            directory: base directory
            progress_bar: TKHelper progress bar or None
            enable_spell_checker: True for spell checker. NOT CURRENTLY SUPPORTED
            show_message: True to show message boxes to user on errors
            exit_on_error:  True to exit on serious errors
            languages_list_dct: dictionary containing the ISO-2 languages we want to load from alternateNames
            feature_code_list_dct: dictionary containing the Geonames.org feature codes we want to load
            supported_countries_dct: dictionary containing the ISO-2 countries we want to load
        """
        self.logger = logging.getLogger(__name__)
        self.geodb = None
        self.show_message = show_message
        self.exit_on_error = exit_on_error
        self.required_db_version = 2
        # Message to user upgrading from DB version earlier than current (2)
        self.db_upgrade_text = 'Adding support for non-English output'
        self.directory: str = directory
        self.progress_bar = progress_bar
        self.line_num = 0
        self.cache_changed: bool = False
        sub_dir = GeoUtil.get_cache_directory(self.directory)
        self.country = None
        self.enable_spell_checker: bool = enable_spell_checker
        self.languages_list_dct = languages_list_dct
        self.feature_code_list_dct = feature_code_list_dct
        self.supported_countries_dct = supported_countries_dct
        self.lang_list = []

        for item in self.languages_list_dct:
            self.lang_list.append(item)

        """
        REMOVED spell check.  Although it works and finds some additional locations it significantly reduces performance.
        if self.enable_spell_checker:
            self.spellcheck:SpellCheck.SpellCheck = SpellCheck.SpellCheck(progress=self.progress_bar,directory=sub_dir,
                                                    countries_dict=self.supported_countries_dct)
        else:
        """
        self.spellcheck = None

        if not os.path.exists(sub_dir):
            self.logger.warning(f'Directory] {sub_dir} NOT FOUND')
            if self.show_message:
                messagebox.showwarning('Folder not found', f'Directory\n\n {sub_dir}\n\n NOT FOUND')
            if exit_on_error:
                sys.exit()

        # Read in Text Replacement dictionary pickle - this has output text replacements
        self.output_replace_cd = CachedDictionary.CachedDictionary(sub_dir, "output_list.pkl")
        self.output_replace_cd.read()
        self.output_replace_dct: Dict[str, str] = self.output_replace_cd.dict
        self.output_replace_list = []

        for item in self.output_replace_dct:
            self.output_replace_list.append(item)
            self.logger.debug(f'Output replace [{item}]')

        self.entry_place = Loc.Loc()

        # Support for Geonames AlternateNames file.  Adds alternate names for entries
        self.alternate_names = AlternateNames.AlternateNames(directory=self.directory, geo_files=self,
                                                             progress_bar=self.progress_bar, filename='alternateNamesV2.txt',
                                                             lang_list=self.lang_list)

    def open_geodb(self, repair_database: bool, query_limit:int) -> bool:
        """
         Open Geoname DB file - this is the db of geoname.org city files and is stored in cache directory under geonames_data.
         The db only contains important fields and only for supported countries.
         If the db doesn't exist and repair flag is True, read the geonames.org files and build DB.   
         The DB has a version table for the schema version.  If the schema changes, the version should be updated.   
         This will check DB schema version and rebuild DB if version is out of date.   
        # Args:   
            repair_database: If True, rebuild database if error or missing   
        Returns:   
            True if error   
        """

        # Use db if it exists and has data and is correct version
        cache_dir = GeoUtil.get_cache_directory(self.directory)
        db_path = os.path.join(cache_dir, 'geodata.db')

        self.logger.debug(f'path for geodata.db: {db_path}')
        err_msg = ''

        # Validate Database setup
        if os.path.exists(db_path):
            # DB was Found
            self.logger.debug(f'DB found at {db_path}')
            self.geodb = GeoDB.GeoDB(db_path=db_path, spellcheck=self.spellcheck,
                                     show_message=self.show_message, exit_on_error=self.exit_on_error,
                                     set_speed_pragmas=True, db_limit=query_limit)

            # Make sure DB is correct version
            ver = self.geodb.get_db_version()
            if ver != self.required_db_version:
                # Wrong version
                err_msg = f'Database version will be upgraded:\n\n{self.db_upgrade_text}\n\n' \
                    f'Upgrading database from V{ver} to V{self.required_db_version}.'
            else:
                # Correct Version.  Make sure DB has reasonable number of records
                count = self.geodb.get_row_count()
                self.logger.info(f'Geoname entries = {count:,}')
                if count < DB_MINIMUM_RECORDS:
                    # Error if DB has under 1000 records
                    err_msg = f'Geoname Database is too small.\n\n {db_path}\n\nRebuilding DB '
        else:
            err_msg = f'Database not found at\n\n{db_path}.\n\nBuilding DB'

        self.logger.debug(f'{err_msg}')
        if err_msg == '':
            # No DB errors detected
            pass
        else:
            # DB error detected - rebuild database if flag set
            if self.show_message:
                messagebox.showinfo('Database Error', err_msg)
            self.logger.debug(err_msg)

            if repair_database:
                if os.path.exists(db_path):
                    self.geodb.close()
                    os.remove(db_path)
                    self.logger.debug('Database deleted')

                self.geodb = GeoDB.GeoDB(db_path=db_path, spellcheck=self.spellcheck,
                                         show_message=self.show_message, exit_on_error=self.exit_on_error,
                                         set_speed_pragmas=True, db_limit=query_limit)
                self.create_geonames_database()

        return False

    def create_geonames_database(self)->bool:
        """
        Create geonames database from geonames.org files.
        You must call self.geodb = GeoDB.GeoDB(...) before this
        # Returns:
            True if error
        """
        # DB didnt exist.  Create tables.
        if self.geodb is None:
            self.logger.error('Cannot create DB: geodb is None')
            return True

        self.geodb.create_tables()
        self.geodb.insert_version(self.required_db_version)

        self.country = Country.Country(progress=self.progress_bar, geo_files=self, lang_list=self.lang_list)

        file_count = 0

        # Set DB version to -1 as invalid until build is completed
        self.geodb.insert_version(-1)

        # Put in country names
        self.country.add_country_names_to_db(geodb=self.geodb)

        # Put in historic names
        self.country.add_historic_names_to_db(self.geodb)

        start_time = time.time()

        # Put in data from geonames.org files
        for fname in ['allCountries.txt', 'ca.txt', 'gb.txt', 'de.txt', 'fr.txt', 'nl.txt']:
            # Read  geoname files
            error = self._add_geoname_file_to_db(fname)  # Read in info (lat/long) for all places from

            if error:
                self.logger.error(f'Error reading geoname file {fname}')
            else:
                file_count += 1

        if file_count == 0:
            self.logger.error(f'No geonames files found in {os.path.join(self.directory, "*.txt")}')
            return True

        self.logger.info(f'Geonames.org files done.  Elapsed ={time.time() - start_time}')

        # Put in geonames.org alternate names
        start_time = time.time()
        err = self.alternate_names.add_alternate_names_to_db()
        if err:
            self.logger.warning(f'Error reading Alternate names.')
        else:
            self.logger.info(f'Alternate names done.  Elapsed ={time.time() - start_time}')
        self.logger.info(f'Geonames entries = {self.geodb.get_row_count():,}')

        # Add aliases
        Normalize.add_aliases_to_database(self)

        # Create Indices
        start_time = time.time()
        self.progress("3) Final Step: Creating Indices for Database...", 95)
        self.geodb.create_geoid_index()
        self.geodb.create_indices()
        self.logger.debug(f'Indices done.  Elapsed ={time.time() - start_time}')

        # Set Database Version
        self.geodb.insert_version(self.required_db_version)
        return err

    def _add_geoname_file_to_db(self, file) -> bool:
        """
            Read in geonames files and build lookup structure

            Read a geoname.org places file and create a db of all the places.
            1. The db contains: Name, Lat, Long, district1ID (State or Province ID),
            district2_id, feat_code

            2. Since Geonames supports over 25M entries, the db is filtered to only the countries and feature types we want
        # Args:
            file: filename (not path) for DB file.  Will be placed in directory from init call

        Returns:
            True if error
        """
        Geofile_row = namedtuple('Geofile_row',
                                 'id name name_asc alt lat lon feat_class feat_code iso iso2 admin1_id'
                                 ' admin2_id admin3_id admin4_id pop elev dem timezone mod')
        self.line_num = 0
        self.progress("Reading {}...".format(file), 0)
        path = os.path.join(self.directory, file)

        if os.path.exists(path):
            fsize = os.path.getsize(path)
            bytes_per_line = 128  # Approximate bytes per line for progress indicator
            with open(path, 'r', newline="", encoding='utf-8', errors='replace') as geofile:
                self.progress("Building Database from {}".format(file), 2)  # initialize progress bar
                reader = csv.reader(geofile, delimiter='\t')
                self.geodb.db.begin()

                # Map line from csv reader into GeonameData namedtuple
                for line in reader:
                    self.line_num += 1
                    if self.line_num % 20000 == 0:
                        # Periodically update progress
                        prog = self.line_num * bytes_per_line * 100 / fsize
                        self.progress(msg=f"1) Building Database from {file}            {prog:.1f}%", val=prog)
                    try:
                        geoname_row = Geofile_row._make(line)
                    except TypeError:
                        self.logger.error(f'Unable to parse geoname location info in {file}  line {self.line_num}')
                        continue

                    # Only handle line if it's  for a country we follow and its
                    # for a Feature tag we're interested in
                    if geoname_row.iso.lower() in self.supported_countries_dct and \
                            geoname_row.feat_code in self.feature_code_list_dct:
                        self.insert_georow(geoname_row)
                        if geoname_row.name.lower() != Normalize.normalize(geoname_row.name, remove_commas=True):
                            self.geodb.insert_alternate_name(geoname_row.name,
                                                             geoname_row.id, 'ut8')

                    if self.progress_bar is not None:
                        if self.progress_bar.shutdown_requested:
                            # Abort DB build.  Clear out partial DB
                            self.geodb.clear_geoname_data()

            self.progress("Write Database", 90)
            self.geodb.db.commit()
            self.progress("Database created", 100)
            return False
        else:
            return True

    @staticmethod
    def _update_geo_row_name(geo_row, name):
        """
            Update the name entry and soundex entry with a name
        #Args:
            geo_row:
            name:
        """
        geo_row[GeoDB.Entry.NAME] = Normalize.normalize(name, remove_commas=True)
        geo_row[GeoDB.Entry.SDX] = GeoUtil.get_soundex(geo_row[GeoDB.Entry.NAME])

    def insert_georow(self, geoname_row):
        """
            Create a Geo_row and insert it into DB
            This also creates new Feature codes PP1M, P1HK, and P10K using population data
        Args:
            geoname_row: ('paris', 'fr', '07', '012', 12.345, 45.123, 'PPL', '34124')

        Returns:
            None
        """
        geo_row = [None] * GeoDB.Entry.MAX
        self._update_geo_row_name(geo_row=geo_row, name=geoname_row.name)

        geo_row[GeoDB.Entry.ISO] = geoname_row.iso.lower()
        geo_row[GeoDB.Entry.ADM1] = geoname_row.admin1_id
        geo_row[GeoDB.Entry.ADM2] = geoname_row.admin2_id
        geo_row[GeoDB.Entry.LAT] = geoname_row.lat
        geo_row[GeoDB.Entry.LON] = geoname_row.lon
        geo_row[GeoDB.Entry.FEAT] = geoname_row.feat_code
        geo_row[GeoDB.Entry.ID] = geoname_row.id

        # Create new Feature codes based on city population
        if int(geoname_row.pop) > 1000000 and 'PP' in geoname_row.feat_code:
            geo_row[GeoDB.Entry.FEAT] = 'PP1M'
        elif int(geoname_row.pop) > 100000 and 'PP' in geoname_row.feat_code:
            geo_row[GeoDB.Entry.FEAT] = 'P1HK'
        elif int(geoname_row.pop) > 10000 and 'PP' in geoname_row.feat_code:
            geo_row[GeoDB.Entry.FEAT] = 'P10K'

        self.geodb.insert(geo_row=geo_row, feat_code=geoname_row.feat_code)

        # Also add abbreviations for USA states
        if geo_row[GeoDB.Entry.ISO] == 'us' and geoname_row.feat_code == 'ADM1':
            # geo_row[GeoDB.Entry.NAME] = geo_row[GeoDB.Entry.ADM1].lower()
            self._update_geo_row_name(geo_row=geo_row, name=geo_row[GeoDB.Entry.ADM1])
            self.geodb.insert(geo_row=geo_row, feat_code=geoname_row.feat_code)

    def get_supported_countries(self) -> [str, int]:
        """ Convert list of supported countries into sorted string """
        nm_msg = ""
        for ky in self.supported_countries_dct:
            nm_msg += ky.upper() + ', '
        return nm_msg, len(self.supported_countries_dct)

    def progress(self, msg, val):
        """ Update progress bar if there is one """
        if self.progress_bar is not None:
            self.progress_bar.update_progress(val, msg)

        # If we're past 80% log item as info, otherwise log as debug
        if val > 80:
            self.logger.info(f'{val:.1f}%  {msg}')
        else:
            self.logger.debug(f'{val:.1f}%  {msg}')

    def close(self):
        if self.geodb:
            self.geodb.close()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  Copyright (c) 2019.       Mike Herbert
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
"""
Provide functions to normalize text strings by converting to lowercase, removing noisewords.   
This is used by the lookup functions and the database build functions.
"""
import re

import unidecode

from geodata import GeodataBuild, GeoUtil, Loc, GeoDB


def normalize_for_scoring(text: str, iso: str) -> str:
    """
        Normalize the title we use to determine how close a match we got. 
        See normalize() for details  
        Also remove noise words such as City Of

    #Args:
        text: text to normalize
        iso: ISO country code

    #Returns:

    """
    text = normalize_for_search(text, iso)
    text = _remove_noise_words(text)
    text = re.sub(', ', ',', text)
    return text


def normalize_for_search(text: str, iso) -> str:
    text = normalize(text=text, remove_commas=False)

    return text


def normalize(text: str, remove_commas: bool) -> str:
    """
    Normalize text - Convert from UTF-8 to lowercase ascii.  Remove most punctuation. Remove commas if parameter set.   
    Then call _phrase_normalize() which normalizes common phrases with multiple spellings, such as saint to st   
    #Args:   
        text:  Text to normalize   
        remove_commas:   True if commas should be removed   

    #Returns:   
        Normalized text

    """

    # Convert UT8 to ascii
    text = unidecode.unidecode(text)
    text = str(text).lower()

    # remove most punctuation
    if remove_commas:
        text = re.sub(r"[^a-zA-Z0-9 $*']+", " ", text)
    else:
        text = re.sub(r"[^a-zA-Z0-9 $*,']+", " ", text)

    text = _phrase_normalize(text)
    return text.strip(' ')


def _phrase_normalize(text: str) -> str:
    """ Strip spaces and normalize spelling for items such as Saint and County """
    # Replacement patterns to clean up entries
    text = re.sub('r.k. |r k ', 'rooms katholieke ', text)
    text = re.sub('sveti |saints |sainte |sint |saint |sankt |st. ', 'st ', text)  # Normalize Saint to St
    text = re.sub(r' co\.', ' county', text)  # Normalize County
    text = re.sub(r'united states of america', 'usa', text)  # Normalize to USA
    text = re.sub(r'united states', 'usa', text)  # Normalize to USA
    text = re.sub(r'town of ', '', text)  # - remove town of
    text = re.sub(r'city of ', '', text)  # - remove city of
    text = re.sub(r'near ', ' ', text)  # - remove near
    text = re.sub(r'cimetiere', 'cemetery', text)   # cimetière

    if 'amt' not in text:
        text = re.sub(r'^mt ', 'mount ', text)

    text = re.sub('  +', ' ', text)  # Strip multiple space
    text = re.sub('county of ([^,]+)', r'\g<1> county', text)  # Normalize 'Township of X' to 'X Township'
    text = re.sub('township of ([^,]+)', r'\g<1> township', text)  # Normalize 'Township of X' to 'X Township'

    text = re.sub('cathedral of ([^,]+)', r'\g<1> cathedral', text)  # Normalize 'Township of X' to 'X Township'
    text = re.sub('palace of ([^,]+)', r'\g<1> palace', text)  # Normalize 'Township of X' to 'X Township'
    text = re.sub('castle of ([^,]+)', r'\g<1> castle', text)  # Normalize 'Township of X' to 'X Township'

    text = re.sub(r',castle', ' castle', text)  # - castle  - remove extra comma
    text = re.sub(r',palace', ' palace', text)  # - palace  - remove extra comma
    text = re.sub(r',cathedral', ' cathedral', text)  # - palace  - remove extra comma

    #text = re.sub(r'battle of ', 'battle of ,', text)  # - battle of - add comma
    text = re.sub(r'k\. at ', ' ', text)  # - killed at  - remove extra comma
    text = re.sub(r'killed ', ' ', text)  # - killed at  - remove extra comma
    return text


def _remove_noise_words(text: str):
    # Calculate score with noise word removal / normalization
    # inp = re.sub('shire', '', inp)

    # Clean up odd case for Normandy American cemetery having only English spelling of Normandy in France
    text = re.sub(r"normandy american ", 'normandie american ', text)
    text = re.sub(r'nouveau brunswick', ' ', text)
    text = re.sub(r'westphalia', 'westfalen', text)
    text = re.sub(r'city of ', ' ', text)
    text = re.sub(r'citta metropolitana di ', ' ', text)
    text = re.sub(r'town of ', ' ', text)
    text = re.sub(r'kommune', '', text)
    text = re.sub(r"politischer bezirk ", ' ', text)
    text = re.sub(r'erry', 'ury', text)
    text = re.sub(r'ery', 'ury', text)
    text = re.sub(r'borg', 'burg', text)
    text = re.sub(r'bourg', 'burg', text)
    text = re.sub(r'urgh', 'urg', text)
    text = re.sub(r'mound', 'mund', text)
    text = re.sub(r'ourne', 'orn', text)
    text = re.sub(r'ney', 'ny', text)
    text = re.sub(r' de ', ' ', text)
    text = re.sub(r' di ', ' ', text)
    text = re.sub(r' du ', ' ', text)
    text = re.sub(r' of ', ' ', text)
    text = re.sub(r' departement', ' ', text)
    text = re.sub(r'battle of ', 'battle of ,', text)  # - battle of - add comma
    text = re.sub(r'k\. at ', ' ', text)  # - killed at  - remove extra comma
    text = re.sub(r'killed ', ' ', text)  # - killed at  - remove extra comma
    text = re.sub(r'royal borough of windsor and maidenhead', 'berkshire', text)

    return text


def remove_aliase(input_words, res_words) -> (str, str):
    if "middlesex" in input_words and "greater london" in res_words:
        input_words = re.sub('middlesex', '', input_words)
        res_words = re.sub('greater london', '', res_words)
    return input_words, res_words


alias_list = {
    'norge': ('norway', '', 'ADM0'),
    'sverige': ('sweden', '', 'ADM0'),
    'osterreich': ('austria', '', 'ADM0'),
    'belgie': ('belgium', '', 'ADM0'),
    'brasil': ('brazil', '', 'ADM0'),
    'danmark': ('denmark', '', 'ADM0'),
    'eire': ('ireland', '', 'ADM0'),
    'magyarorszag': ('hungary', '', 'ADM0'),
    'italia': ('italy', '', 'ADM0'),
    'espana': ('spain', '', 'ADM0'),
    'deutschland': ('germany', '', 'ADM0'),
    'prussia': ('germany', '', 'ADM0'),
    'suisse': ('switzerland', '', 'ADM0'),
    'schweiz': ('switzerland', '', 'ADM0'),

    'bayern': ('bavaria', 'de', 'ADM1'),
    'westphalia': ('westfalen', 'de', 'ADM1'),

    'normandy': ('normandie', 'fr', 'ADM1'),
    'brittany': ('bretagne', 'fr', 'ADM1'),
    'burgundy': ('bourgogne franche comte', 'fr', 'ADM1'),
    'franche comte': ('bourgogne franche comte', 'fr', 'ADM1'),
    'aquitaine': ('nouvelle aquitaine', 'fr', 'ADM1'),
    'limousin': ('nouvelle aquitaine', 'fr', 'ADM1'),
    'poitou charentes': ('nouvelle aquitaine', 'fr', 'ADM1'),
    'alsace': ('grand est', 'fr', 'ADM1'),
    'champagne ardenne': ('grand est', 'fr', 'ADM1'),
    'lorraine': ('grand est', 'fr', 'ADM1'),
    'languedoc roussillon': ('occitanie', 'fr', 'ADM1'),
    'nord pas de calais': ('hauts de france', 'fr', 'ADM1'),
    'picardy': ('hauts de france', 'fr', 'ADM1'),
    'auvergne': ('auvergne rhone alpes', 'fr', 'ADM1'),
    'rhone alpes': ('auvergne rhone alpes', 'fr', 'ADM1'),

    'breconshire': ('sir powys', 'gb', 'ADM2'),
    }


def add_aliases_to_database(geo_files: GeodataBuild):
    #  Add alias names to DB
    place = Loc.Loc()
    for ky in alias_list:
        row = alias_list.get(ky)
        place.clear()

        # Create Geo_row
        # ('paris', 'fr', '07', '012', '12.345', '45.123', 'PPL')
        geo_row = [None] * GeoDB.Entry.MAX
        geo_row[GeoDB.Entry.FEAT] = row[2]
        geo_row[GeoDB.Entry.ISO] = row[1].lower()
        geo_row[GeoDB.Entry.LAT] = '99.9'
        geo_row[GeoDB.Entry.LON] = '99.9'
        geo_row[GeoDB.Entry.ADM1] = ''
        geo_row[GeoDB.Entry.ADM2] = ''

        geo_files._update_geo_row_name(geo_row=geo_row, name=ky)
        if row[2] == 'ADM1':
            geo_row[GeoDB.Entry.ADM1] = ky
            place.place_type = Loc.PlaceType.ADMIN1
        elif row[2] == 'ADM2':
            geo_row[GeoDB.Entry.ADM2] = ky
            place.place_type = Loc.PlaceType.ADMIN2
        else:
            place.place_type = Loc.PlaceType.COUNTRY
            place.country_name = ky

        place.country_iso = row[1]
        place.admin1_name = geo_row[GeoDB.Entry.ADM1]
        place.admin2_name = geo_row[GeoDB.Entry.ADM2]

        # Lookup main entry and get GEOID
        geo_files.geodb.lookup_place(place)
        if place.result_type in GeoUtil.successful_match and len(place.georow_list) > 0:
            geo_files.geodb.copy_georow_to_place(row=place.georow_list[0], place=place)
            # place.format_full_nm(geodata.geo_files.output_replace_dct)

        geo_row[GeoDB.Entry.ID] = place.geoid

        geo_files.geodb.insert(geo_row=geo_row, feat_code=row[2])


def admin1_normalize(admin1_name: str, iso):
    """ Normalize historic or colloquial Admin1 names to standard """
    admin1_name = normalize_for_search(admin1_name, iso)
    if iso == 'de':
        admin1_name = re.sub(r'bayern', 'bavaria', admin1_name)
        admin1_name = re.sub(r'westphalia', 'westfalen', admin1_name)

    if iso == 'fr':
        admin1_name = re.sub(r'normandy', 'normandie', admin1_name)
        admin1_name = re.sub(r'brittany', 'bretagne', admin1_name)

        admin1_name = re.sub(r'burgundy', 'bourgogne franche comte', admin1_name)
        admin1_name = re.sub(r'franche comte', 'bourgogne franche comte', admin1_name)
        admin1_name = re.sub(r'aquitaine', 'nouvelle aquitaine', admin1_name)
        admin1_name = re.sub(r'limousin', 'nouvelle aquitaine', admin1_name)
        admin1_name = re.sub(r'poitou charentes', 'nouvelle aquitaine', admin1_name)
        admin1_name = re.sub(r'alsace', 'grand est', admin1_name)
        admin1_name = re.sub(r'champagne ardenne', 'grand est', admin1_name)
        admin1_name = re.sub(r'lorraine', 'grand est', admin1_name)
        admin1_name = re.sub(r'languedoc roussillon', 'occitanie', admin1_name)
        admin1_name = re.sub(r'midi pyrenees', 'occitanie', admin1_name)
        admin1_name = re.sub(r'nord pas de calais', 'hauts de france', admin1_name)
        admin1_name = re.sub(r'picardy', 'hauts de france', admin1_name)
        admin1_name = re.sub(r'auvergne', 'auvergne rhone alpes', admin1_name)
        admin1_name = re.sub(r'rhone alpes', 'auvergne rhone alpes', admin1_name)

    return admin1_name


def admin2_normalize(admin2_name: str, iso) -> (str, bool):
    """
        Normalize historic or colloquial Admin2 names to standard

    Args:
        admin2_name: 
        iso: 

    Returns: (result, modified) - result is new string, modified - True if modified

    """
    admin2_name = normalize_for_search(admin2_name, iso)

    mod = False

    if iso == 'gb':
        # res = re.sub(r'middlesex', ' ', res)
        admin2_name = re.sub(r'breconshire', 'sir powys', admin2_name)
        mod = True

    return admin2_name, mod


def country_normalize(country_name) -> (str, bool):
    """
    normalize local language Country name to standardized English country name for lookups
    :param country_name:
    :return: (result, modified)
    result - new string
    modified - True if modified
    """
    country_name = re.sub(r'\.', '', country_name)  # remove .

    natural_names = {
        'norge': 'norway',
        'sverige': 'sweden',
        'osterreich': 'austria',
        'belgie': 'belgium',
        'brasil': 'brazil',
        'danmark': 'denmark',
        'magyarorszag': 'hungary',
        'italia': 'italy',
        'espana': 'spain',
        'deutschland': 'germany',
        'prussia': 'germany',
        'suisse': 'switzerland',
        'schweiz': 'switzerland',

        }
    if natural_names.get(country_name):
        country_name = natural_names.get(country_name)
        return country_name.strip(' '), True
    else:
        return country_name.strip(' '), False



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  Copyright (c) 2019.       Mike Herbert
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA

# The tab separated columns in geoname.org file rows are as follows
"""
Add alternate names for places using the Geonames.org Alternate names file 
"""
from geodata import GeodataBuild, Loc, GeoUtil, GeoDB, FileReader

ALT_GEOID = 1
ALT_LANG = 2
ALT_NAME = 3


class AlternateNames(FileReader.FileReader):
    """
    Read in Geonames.org Alternate names V2 file and add appropriate entries to the altnames table.
    Each row in file contains a geoname ID, an alternative name for that entity, and the language.
    If the lang is in lang_list and the ID is ALREADY in our geonames dictionary, we add this as an alternative name
    """

    def __init__(self, directory: str, filename: str, progress_bar, geo_files: GeodataBuild, lang_list):
        """
            Read in geonames alternate names file and add to geodata database in alt_names table
        # Args:
            directory: base directory for alternate names file
            filename: filename of geonames alternate_namesV2.txt file
            progress_bar: tkhelper progress bar or None
            geo_files: GeodataFiles instance
            lang_list: List of ISO languages we want to support, e.g. ['fr', 'es']
        """
        super().__init__(directory, filename, progress_bar)
        self.sub_dir = GeoUtil.get_cache_directory(directory)
        self.geo_files: GeodataBuild.GeodataBuild = geo_files
        self.lang_list = lang_list
        self.loc = Loc.Loc()

    def add_alternate_names_to_db(self) -> bool:
        """
        Read alternate names file into database
        # Returns:
            True if error
        """
        self.geo_files.geodb.db.begin()
        # Read in file.  This will call handle_line for each line in file
        res = super().read()
        self.geo_files.geodb.db.commit()
        return res

    def handle_line(self, line_num, row):
        """
        For each line in file, add item to alternate name DB if we support that language
        Also add to main DB if lang is not English and item is not an ADM item
        :param line_num:  file line number
        :param row: line in file to be handled
        :return: None
        """
        alt_tokens = row.split('\t')
        if len(alt_tokens) != 10:
            self.logger.debug(f'Incorrect number of tokens: {alt_tokens} line {line_num}')
            return

        self.loc.georow_list = []
        if alt_tokens[ALT_LANG] == '':
            alt_tokens[ALT_LANG] = 'en'
            
        # Alternate names are in multiple languages.  Only add if item is in requested lang list
        if alt_tokens[ALT_LANG] in self.lang_list:
            # Only Add this alias if  DB already has an entry (since geoname DB is filtered based on feature)
            if 'estminser' in alt_tokens[ALT_NAME]:
                pass
 
            # See if item has an entry with same GEOID in Main DB
            dbid:str = str(self.geo_files.geodb.geoid_main_dict.get(alt_tokens[ALT_GEOID]))
            if dbid is not None:
                self.loc.target = dbid
                # Retrieve entry
                self.geo_files.geodb.lookup_main_dbid(place=self.loc)
            else:
                # See if item has an entry with same GEOID in Admin DB
                dbid = self.geo_files.geodb.geoid_admin_dict.get(alt_tokens[ALT_GEOID])
                if dbid is not None:
                    self.loc.target = dbid
                    # Retrieve entry
                    self.geo_files.geodb.lookup_admin_dbid(place=self.loc)

            if len(self.loc.georow_list) > 0:
                # Create an entry in the alternate name DB  with this name and soundex

                # Convert row to list. modify name and soundex and add to alternate name DB
                # Update the name in the new row with the alternate name
                update = list(self.loc.georow_list[0])

                # Make sure this entry has a different name from existing entry
                if update[ALT_NAME] != alt_tokens[ALT_NAME]:
                    self.geo_files._update_geo_row_name(geo_row=update, name=alt_tokens[ALT_NAME])
                    new_row = tuple(update)  # Convert back to tuple

                    if 'ADM1' not in update[GeoDB.Entry.FEAT] and 'ADM2' not in update[GeoDB.Entry.FEAT]:
                        #  Add to main DB if not English or not ADM1/ADM2
                        self.geo_files.geodb.insert(geo_row=new_row, feat_code=update[GeoDB.Entry.FEAT])
                        self.count += 1

                    # Add name to altnames table
                    if alt_tokens[ALT_LANG] != 'en':
                        self.geo_files.geodb.insert_alternate_name(alt_tokens[ALT_NAME],
                                                                   alt_tokens[ALT_GEOID], alt_tokens[ALT_LANG])
                        self.count += 1


    def cancel(self):
        """
        User requested cancel of database build.
        Quit DB build.
        :return: None
        """
        self.geo_files.geodb.db.commit()
        # self.geo_files.geodb.clear_geoname_data()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  Copyright (c) 2019.       Mike Herbert
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA

""" Build database queries for various lookups """
import re

from geodata import Loc
from geodata.GeoUtil import Query, Result, get_soundex


class Typ:
    CITY = 0
    ADMIN2 = 1
    ADMIN1 = 2
    COUNTRY = 3
    ADMIN1_ID = 4
    ADMIN2_ID = 5
    ADMIN1_ALT_NAME = 6
    GEOID = 7


class QueryList:
    @staticmethod
    def build_query_list(typ: int, query_list, place: Loc):
        """
        
        Args:
            typ: 
            query_list: 
            place: 

        Returns:

        """
        if typ == Typ.CITY:
            QueryList.query_list_city(query_list, place)
        elif typ == Typ.ADMIN2:
            QueryList.query_list_admin2(query_list, place)
        elif typ == Typ.ADMIN1:
            QueryList.query_list_admin1(query_list, place)
        elif typ == Typ.COUNTRY:
            QueryList.query_list_country(query_list, place)
        elif typ == Typ.ADMIN1_ID:
            QueryList.query_list_admin1_id(query_list, place)
        elif typ == Typ.ADMIN2_ID:
            QueryList.query_list_admin2_id(query_list, place)
        elif typ == Typ.ADMIN1_ALT_NAME:
            QueryList.query_list_admin1_alt_name(query_list, place)
        elif typ == Typ.GEOID:
            QueryList.query_list_geoid(query_list, place)

    @staticmethod
    def query_list_city(query_list, place: Loc):
        """
        Search for  city - try the most exact match, then less exact matches
        """
        sdx = get_soundex(place.target)
        pattern = QueryList.create_wildcard(place.target)
        if len(place.country_iso) == 0:
            # NO COUNTRY - try lookup by name.
            if place.target in pattern:
                query_list.append(Query(where="name = ?",
                                        args=(place.target,),
                                        result=Result.PARTIAL_MATCH))
            # lookup by wildcard name
            if '*' in place.target:
                query_list.clear()
                query_list.append(Query(where="name LIKE ?",
                                        args=(pattern,),
                                        result=Result.WILDCARD_MATCH))
            else:
                query_list.append(Query(where="name LIKE ?",
                                        args=(pattern,),
                                        result=Result.WORD_MATCH))

            # lookup by soundex
            query_list.append(Query(where="sdx = ?",
                                    args=(sdx,),
                                    result=Result.SOUNDEX_MATCH))
            return

        if len(place.admin1_name) > 0:
            # lookup by name, ADMIN1, country
            if place.target in pattern:
                query_list.append(Query(
                    where="name = ? AND country = ? AND admin1_id = ?",
                    args=(place.target, place.country_iso, place.admin1_id),
                    result=Result.STRONG_MATCH))

            # add lookup by wildcard name, ADMIN1, country
            if '*' in place.target:
                query_list.clear()

                query_list.append(Query(
                    where="name LIKE ? AND country = ? AND admin1_id = ?",
                    args=(pattern, place.country_iso, place.admin1_id),
                    result=Result.WILDCARD_MATCH))
            else:
                query_list.append(Query(
                    where="name LIKE ? AND country = ? AND admin1_id = ?",
                    args=(pattern, place.country_iso, place.admin1_id),
                    result=Result.WORD_MATCH))
        else:
            # add lookup by wildcard  name, country
            if '*' in place.target:
                query_list.clear()

                query_list.append(Query(where="name LIKE ? AND country = ?",
                                        args=(pattern, place.country_iso),
                                        result=Result.WILDCARD_MATCH))
            else:
                query_list.append(Query(where="name LIKE ? AND country = ?",
                                        args=(pattern, place.country_iso),
                                        result=Result.WORD_MATCH))

        # add lookup by Soundex , country
        query_list.append(Query(where="sdx = ? AND country = ?",
                                args=(sdx, place.country_iso),
                                result=Result.SOUNDEX_MATCH))

    @staticmethod
    def query_list_admin2(query_list, place: Loc):
        # Try Admin queries and find best match
        # sdx = get_soundex(place.target)
        pattern = QueryList.create_wildcard(place.target)

        if len(place.country_iso) == 0:
            query_list.append(Query(where="name = ?  AND f_code=?",
                                    args=(place.target, 'ADM2'),
                                    result=Result.PARTIAL_MATCH))
        else:
            query_list.append(Query(where="name = ? AND country = ? AND f_code=?",
                                    args=(place.target, place.country_iso, 'ADM2'),
                                    result=Result.PARTIAL_MATCH))

        query_list.append(Query(where="name LIKE ? AND country = ? AND f_code=?",
                                args=(place.target, place.country_iso, 'ADM2'),
                                result=Result.PARTIAL_MATCH))

        if '*' in place.target:
            query_list.clear()
            query_list.append(Query(where="name LIKE ? AND country = ? AND f_code=?",
                                    args=(pattern, place.country_iso, 'ADM2'),
                                    result=Result.WILDCARD_MATCH))
        else:
            query_list.append(Query(
                where="name LIKE ? AND country = ? AND admin1_id = ?",
                args=(pattern, place.country_iso, place.admin1_id),
                result=Result.WORD_MATCH))

    @staticmethod
    def query_list_admin1(query_list, place: Loc):
        """Search for Admin1 entry"""
        sdx = get_soundex(place.target)
        pattern = QueryList.create_wildcard(place.target)
        query_list.append(Query(where="name = ? AND country = ? AND f_code = ? ",
                                args=(place.target, place.country_iso, 'ADM1'),
                                result=Result.STRONG_MATCH))
        if '*' in place.target:
            query_list.clear()

            query_list.append(Query(where="name LIKE ? AND country = ?  AND f_code = ?",
                                    args=(pattern, place.country_iso, 'ADM1'),
                                    result=Result.WILDCARD_MATCH))
        else:
            query_list.append(Query(where="name LIKE ? AND country = ?  AND f_code = ?",
                                    args=(pattern, place.country_iso, 'ADM1'),
                                    result=Result.WORD_MATCH))

        query_list.append(Query(where="sdx = ? AND country = ? AND f_code=?",
                                args=(sdx, place.country_iso, 'ADM1'),
                                result=Result.SOUNDEX_MATCH))

    @staticmethod
    def query_list_country(query_list, place: Loc):
        """Search for Admin1 entry"""
        sdx = get_soundex(place.target)
        # pattern = QueryList.create_wildcard(place.target)

        query_list.append(where="country = ? AND f_code = ? ",
                          args=(place.country_iso, 'ADM0'),
                          result=Result.STRONG_MATCH)

        query_list.append(where="sdx = ?  AND f_code=?",
                          args=(sdx, 'ADM0'),
                          result=Result.SOUNDEX_MATCH)

    @staticmethod
    def query_list_admin1_id(query_list, place: Loc):
        """Search for Admin1 entry"""
        # sdx = get_soundex(place.target)
        pattern = QueryList.create_wildcard(place.target)

        if place.country_iso == '':
            query_list.append(Query(where="name = ?  AND f_code = ? ",
                                    args=(place.target, 'ADM1'),
                                    result=Result.STRONG_MATCH))

            query_list.append(Query(where="name LIKE ? AND f_code = ?",
                                    args=(pattern, 'ADM1'),
                                    result=Result.WILDCARD_MATCH))
        else:
            query_list.append(Query(where="name = ? AND country = ? AND f_code = ? ",
                                    args=(place.target, place.country_iso, 'ADM1'),
                                    result=Result.STRONG_MATCH))
            query_list.append(Query(where="name LIKE ? AND country = ?  AND f_code = ?",
                                    args=(pattern, place.country_iso, 'ADM1'),
                                    result=Result.WILDCARD_MATCH))

        query_list.append(Query(where="name = ?  AND f_code = ?",
                                args=(place.target, 'ADM1'),
                                result=Result.SOUNDEX_MATCH))

    @staticmethod
    def query_list_admin2_id(query_list, place: Loc):
        """Search for Admin1 entry"""
        # sdx = get_soundex(place.target)
        pattern = QueryList.create_wildcard(place.target)

        if len(place.admin1_id) > 0:
            query_list.append(Query(where="name = ? AND country = ? AND admin1_id=? AND f_code=?",
                                    args=(place.target, place.country_iso, place.admin1_id, 'ADM2'),
                                    result=Result.STRONG_MATCH))
            query_list.append(Query(where="name LIKE ? AND country = ? and admin1_id = ? AND f_code=?",
                                    args=(place.target, place.country_iso, place.admin1_id, 'ADM2'),
                                    result=Result.WILDCARD_MATCH))
            query_list.append(Query(where="name LIKE ? AND country = ? and admin1_id = ? AND f_code=?",
                                    args=(pattern, place.country_iso, place.admin1_id, 'ADM2'),
                                    result=Result.WILDCARD_MATCH))
        else:
            query_list.append(Query(where="name = ? AND country = ? AND f_code=?",
                                    args=(place.target, place.country_iso, 'ADM2'),
                                    result=Result.STRONG_MATCH))
            query_list.append(Query(where="name LIKE ? AND country = ? AND f_code=?",
                                    args=(place.target, place.country_iso, 'ADM2'),
                                    result=Result.WILDCARD_MATCH))
            query_list.append(Query(where="name LIKE ? AND country = ? AND f_code=?",
                                    args=(pattern, place.country_iso, 'ADM2'),
                                    result=Result.WILDCARD_MATCH))

    @staticmethod
    def query_list_admin1_alt_name(query_list, place: Loc):
        """Search for Admin1 entry"""
        query_list.append(where="admin1_id = ? AND country = ?  AND f_code = ? ",
                          args=(place.target, place.country_iso, 'ADM1'),
                          result=Result.STRONG_MATCH)

    @staticmethod
    def query_list_geoid(query_list, place: Loc) -> None:
        """Search for GEOID"""
        query_list.append(where="geoid = ? ",
                          args=(place.target,),
                          result=Result.STRONG_MATCH)

    @staticmethod
    def query_list_admin1_name_direct(query_list, lookup_target, iso):
        """Search for Admin1 entry"""
        query_list.append(where="admin1_id = ? AND country = ?  AND f_code = ? ",
                          args=(lookup_target, iso, 'ADM1'),
                          result=Result.STRONG_MATCH)

    @staticmethod
    def query_list_admin2_name_direct(query_list, lookup_target, iso, admin1_id):
        """Search for Admin2 entry"""
        query_list.append(where="admin2_id = ? AND country = ? AND admin1_id = ?",
                          args=(lookup_target, iso, admin1_id),
                          result=Result.STRONG_MATCH)

        query_list.append(where="admin2_id = ? AND country = ?",
                          args=(lookup_target, iso),
                          result=Result.PARTIAL_MATCH)

    @staticmethod
    def create_wildcard(pattern):
        """

        :param pattern:
        :return:
        """
        # Create SQL wildcard pattern (convert * to %).  Add % on end
        if '*' in pattern:
            return re.sub(r"\*", "%", pattern)
        else:
            return f'%{pattern}%'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  Copyright (c) 2019.       Mike Herbert
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
"""Calculate a heuristic score for how well a result place name matches a target place name."""
import copy
import logging
import re

from geodata import GeoUtil, Loc, Normalize, Geodata


class Score:
    VERY_GOOD = 19
    GOOD = 40
    POOR = 85
    VERY_POOR = 120


class MatchScore:
    """
    Calculate a heuristic score for how well a result place name matches a target place name. The score is based on percent
    of characters that didnt match plus other items - described in match_score()
    """

    def __init__(self):
        self.logger = logging.getLogger(__name__)

        # Weighting for each input term match - prefix, city, adm2, adm1, country
        self.token_weight = [0.0, 1.0, 0.6, 0.8, 0.9, 0, 0, 0]

        # Weighting for each part of score
        self.wildcard_penalty = -41.0
        self.first_token_match_bonus = 40.0
        self.wrong_order_penalty = -3.0
        self.prefix_weight = 1.9

        self.feature_weight = 0.12
        self.result_weight = 0.17  # weight for match score of Result name
        self.input_weight = 1.0 - self.result_weight - self.feature_weight  # Weight for match score of user input name

        # Out weight + Feature weight must be less than 1.0.
        if self.result_weight + self.feature_weight > 1.0:
            self.logger.error('Out weight + Feature weight must be less than 1.0')

        self.in_score = 99.0
        self.out_score = 99.0

    def match_score(self, target_place: Loc, result_place: Loc) -> float:
        """
            Calculate a heuristic score for how well a result place name matches a target place name.  The score is based on
            percent of characters that didnt match in input and output (plus other items described below).
            Mismatch score is 0-100% reflecting the percent mismatch between the user input and the result.  This is then
            adjusted by Feature type (large city gives best score) plus other items to give a final heuristic where
            -10 is perfect match of a large city and 100 is no match.

            A) Heuristic:
            1) Create 5 part title (prefix, city, county, state/province, country)
            2) Normalize text - Normalize.normalize_for_scoring()
            3) Remove sequences over 3 chars that match in target and result
            4) Calculate inscore - percent of characters in input that didn't match output.  Weighted by term (county is lower weight)
                    Exact match of city term gets a bonus
            5) Calculate outscore - percent of characters in output that didn't match input

            B) Score components (All are weighted except Prefix and Parse):
            in_score - percent of characters in input that didnt match output
            out_score - percent of characters in output that didnt match input
            feature_score - Geodata.feature_priority().  adjustment based on Feature type.  More important features (larger city)
            get lower result
            wildcard_penalty - score is raised by 41 if it includes a wildcard
            prefix_penalty -  score is raised by length of Prefix
            parse_penalty - score is raised by 3 if lookup was done with terms out of order

            C) A standard text difference, such as Levenstein, was not used because those treat both strings as equal,
            whereas this treats the User text as more important than DB result text and also weights each token.  A user's
            text might commonly be something like: Paris, France and a DB result of Paris, Paris, Ile De France, France.
            The Levenstein distance would be large, but with this heuristic, the middle terms can have lower weights, and
            having all the input matched can be weighted higher than mismatches on the county and province.  This heuristic gives
            a score of -9 for Paris, France.

        # Args:
            target_place:  Loc  with users entry.
            result_place:  Loc with DB result.
        # Returns:
            score
        """
        target_tkn_len = [0] * 20

        # Create full RESULT title (prefix,city,county,state,country)
        result_place.prefix = ' '
        result_words = result_place.get_five_part_title()
        result_words = Normalize.normalize_for_scoring(result_words, result_place.country_iso)
        result_place.original_entry = copy.copy(result_words)
        res_tokens = result_words.split(',')

        # Create full TARGET  title (prefix,city,county,state,country)
        # Clean up prefix - remove any words that are in city, admin1 or admin2 from Prefix
        target_words = target_place.get_five_part_title()
        target_words = Normalize.normalize_for_scoring(target_words, target_place.country_iso)
        target_tokens = target_words.split(',')
        target_tokens[0] = Loc.Loc.matchscore_prefix(target_tokens[0], result_words)

        target_words, result_words = Normalize.remove_aliase(target_words, result_words)

        # Store length of original tokens in target.  This is used for percent unmatched calculation
        for it, tk in enumerate(target_tokens):
            target_tokens[it] = target_tokens[it].strip(' ')
            target_tkn_len[it] = len(target_tokens[it])

        # Remove sequences that match in target and result
        result_words, target_words = GeoUtil.remove_matching_sequences(text1=result_words, text2=target_words, min_len=3)

        # Calculate score for input match
        self.in_score = self._calculate_input_score(target_tkn_len, target_tokens, target_words, res_tokens)

        # Calculate score for output match
        self.out_score = self._calculate_output_score(result_words, result_place.original_entry)

        if not target_place.standard_parse:
            # If Tokens were not in hierarchical order, give penalty
            parse_penalty = self.wrong_order_penalty
        else:
            parse_penalty = 0.0

        if '*' in target_place.original_entry:
            # if it was a wildcard search it's hard to rank - add a penalty
            wildcard_penalty = self.wildcard_penalty
        else:
            wildcard_penalty = 0.0

        # Prefix penalty for length of prefix
        if target_tkn_len[0] > 0:
            prefix_penalty = 3 + target_tkn_len[0]
        else:
            prefix_penalty = 0

        # Feature score is to ensure "important" places get higher rank (large city, etc)
        feature_score = Geodata.Geodata._feature_priority(result_place.feature)

        # Add up scores - Each item is appx 0-100 and weighted
        score: float = self.in_score * self.input_weight + self.out_score * self.result_weight + feature_score * self.feature_weight + \
                       wildcard_penalty + prefix_penalty * self.prefix_weight + parse_penalty

        #self.logger.debug(f'SCORE {score:.1f} res=[{result_place.original_entry}] pref=[{target_place.prefix}]\n'
        #                  f'inp=[{",".join(target_tokens)}]  outSc={self.out_score * self.result_weight:.1f}% '
        #                  f'inSc={self.in_score * self.input_weight:.1f}% feat={feature_score * self.feature_weight:.1f} {result_place.feature}  '
        #                  f'wild={wildcard_penalty} pref={prefix_penalty * self.prefix_weight:.1f}')

        return score

    def _calculate_input_score(self, inp_len: [], inp_tokens: [], input_words, res_tokens: []) -> float:
        num_inp_tokens = 0.0
        in_score = 0

        # For each input token calculate percent of unmatched size vs original size
        unmatched_input_tokens = input_words.split(',')

        # Each token in place hierarchy gets a different weighting
        #      Prefix, city,county, state, country
        self.score_diags = ''
        unmatched_input_tokens[0] = inp_tokens[0]
        match_bonus = 0

        # Calculate percent of USER INPUT text that was unmatched, then apply weighting
        for idx, tk in enumerate(inp_tokens):
            if inp_len[idx] > 0:
                unmatched_percent = int(100.0 * len(unmatched_input_tokens[idx].strip(' ')) / inp_len[idx])
                in_score += unmatched_percent * self.token_weight[idx]
                self.score_diags += f'  {idx}) [{tk}][{unmatched_input_tokens[idx]}] {unmatched_percent}% * {self.token_weight[idx]} '
                # self.logger.debug(f'{idx}) Rem=[{unmatched_input_tokens[idx].strip(" " )}] wgtd={unmatched_percent * self.weight[idx]}')
                num_inp_tokens += 1.0 * self.token_weight[idx]
                # self.logger.debug(f'{idx} [{inp_tokens2[idx]}:{inp_tokens[idx]}] rawscr={sc}% orig_len={inp_len[idx]} wgt={self.weight[idx]}')
                if idx == 1:
                    # If the full first or second token of the result is in input then improve score
                    # Bonus for a full match as against above partial matches
                    if res_tokens[idx] in inp_tokens[idx]:
                        in_score -= self.first_token_match_bonus
                    # If exact match of term, give bonus
                    if inp_tokens[idx] == res_tokens[idx]:
                        if idx == 0:
                            match_bonus -= 9
                        else:
                            match_bonus -= 3

        # Average over number of tokens (with fractional weight).  Gives 0-100% regardless of weighting and number of tokens
        if num_inp_tokens > 0:
            in_score = in_score / num_inp_tokens
        else:
            in_score = 0

        return in_score + match_bonus + 10

    def _calculate_output_score(self, unmatched_result: str, original_result: str) -> float:
        """
        Calculate score for output (DB result).
        :param unmatched_result: The text of the DB result that didnt match the user's input
        :param original_result: The original DB result
        :return: 0=strong match, 100=no match
        """

        # Remove spaces and commas from original and unmatched result
        original_result = re.sub(r'[ ,]', '', original_result)
        unmatched = re.sub(r'[ ,]', '', unmatched_result)

        orig_res_len = len(original_result)
        if orig_res_len > 0:
            # number of chars of DB RESULT text that matched target - scaled from 0 (20 or more matched) to 100 (0 matched)
            out_score_1 = (20.0 - min((orig_res_len - len(unmatched)), 20.0)) * 5.0
            # self.logger.debug(f'matched {orig_res_len - len(unmatched)} [{unmatched}]')

            # Percent of unmatched
            out_score_2 = 100.0 * len(unmatched) / orig_res_len

            out_score = out_score_1 * 0.1 + out_score_2 * 0.9
        else:
            out_score = 0.0

        self.score_diags += f'\noutrem=[{unmatched}]'

        return out_score

    @staticmethod
    def _adjust_adm_score(score, feat):
        return score



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  Copyright (c) 2019.       Mike Herbert
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
"""    Holds the details about a Location: Name,state/province, etc as well as
    the lookup results """
import argparse
import logging
import re
from typing import List, Tuple

from geodata import GeoUtil, Normalize, ArgumentParserNoExit


# What type of entity is this place?
class PlaceType:
    COUNTRY = 0
    ADMIN1 = 1
    ADMIN2 = 2
    CITY = 3
    PREFIX = 4
    ADVANCED_SEARCH = 5


place_type_name_dict = {
    PlaceType.COUNTRY        : 'Country',
    PlaceType.ADMIN1         : 'STATE/PROVINCE',
    PlaceType.ADMIN2         : 'COUNTY',
    PlaceType.CITY           : ' ',
    PlaceType.ADVANCED_SEARCH: ' '
    }


class Loc:
    """
    Holds the details about a Location: Name, county, state/province, country, lat/long as well as
    the lookup results
    Parses a name into Loc items (county, state, etc)
    """

    def __init__(self):
        """
        Init
        """
        self.logger = logging.getLogger(__name__)
        self.clear()
        self.event_year: int = 0

    def clear(self):
        # Place geo info
        self.original_entry: str = ""
        self.formatted_name: str = ''
        self.lat: float = float('NaN')  # Latitude
        self.lon: float = float('NaN')  # Longitude
        self.country_iso: str = ""  # Country ISO code
        self.country_name: str = ''
        self.city1: str = ""  # City or entity name
        self.admin1_name: str = ""  # Admin1 (State/province/etc)
        self.admin2_name: str = ""  # Admin2 (county)
        self.admin1_id: str = ""  # Admin1 Geoname ID
        self.admin2_id = ""  # Admin2 Geoname ID
        self.prefix: str = ""  # Prefix (entries before city)
        # self.extra: str = ''  # Extra tokens that dont get put in prefix
        self.feature: str = ''  # Geoname feature code
        self.place_type: int = PlaceType.COUNTRY  # Is this a Country , Admin1 ,admin2 or city?
        self.target: str = ''  # Target for lookup
        self.geoid: str = ''  # Geoname GEOID
        self.prefix_commas: str = ''
        self.id = ''
        self.enclosed_by = ''
        self.standard_parse = True
        self.updated_entry = ''
        self.score = 9999

        # Lookup result info
        self.status: str = ""
        self.status_detail: str = ""
        self.result_type: int = GeoUtil.Result.NO_MATCH  # Result type of lookup
        self.result_type_text: str = ''  # Text version of result type
        self.georow_list: List[Tuple] = [()]  # List of items that matched this location

        self.georow_list.clear()

    def parse_place(self, place_name: str, geo_files):
        """
            Given a comma separated place name,   
            parse into its city, admin1, country and type of entity (city, country etc)   
        #Args:   
            place_name: The place name to parse   
            geo_files: Geo_files   
        #Returns:   
            Fields in Loc (city, adm1, adm2, iso) are updated based on parsing. self.status has Result status code   
        """
        self.clear()
        self.original_entry = place_name

        # Convert open-brace and open-paren to comma.  close brace/paren will be stripped by normalize()
        res = re.sub(r'\[', ',', place_name)
        res = re.sub(r'\(', ',', res)

        tokens = res.split(",")
        if len(tokens[-1]) == 0:
            # Last item is blank, so remove it
            tokens = tokens[:-1]

        token_count = len(tokens)
        self.place_type = PlaceType.CITY

        # First, try to parse and validate State/Province, and Country from last two tokens  
        # If one other token, parse as city
        # If two other tokens, parse as city, admin2
        # First two tokens are also copied to prefix.
        # Place type is the leftmost item we found - either City, Admin2, Admin2, or Country
        # If '--' in name, then extract advanced search options 

        if '--' in place_name:
            # Advanced Search - Pull out filter flags if present
            self.logger.debug('filter')
            self.get_filter_parameters(place_name)
            return
        elif token_count > 0:
            #  COUNTRY - right-most token should be country
            self.country_name = Normalize.normalize_for_search(tokens[-1], "")
            self.target = self.country_name

            # Validate country
            self.country_iso = geo_files.geodb.get_country_iso(self)  # Get Country country_iso
            if self.country_iso != '':
                # self.logger.debug(f'Found country. iso = [{self.country_iso}]')
                self.place_type = PlaceType.COUNTRY
            else:
                # Last token is not COUNTRY.
                # Append dummy token  so we now have <tokens>, x
                tokens.append('_')
                token_count = len(tokens)
                self.result_type = GeoUtil.Result.NO_COUNTRY
                self.country_iso = ''
                self.country_name = ''

        if token_count > 1:
            #  See if 2nd to last token is Admin1
            self.admin1_name = Normalize.admin1_normalize(tokens[-2], self.country_iso)

            if len(self.admin1_name) > 0:
                self.target = self.admin1_name
                # Lookup Admin1
                geo_files.geodb.wide_search_admin1_id(self)
                if self.admin1_id != '':
                    # Found Admin1
                    self.place_type = PlaceType.ADMIN1
                else:
                    # Last token is not Admin1 - append dummy token so we have <tokens>, admin1, country
                    self.admin1_name = ''
                    # Add dummy token for admin1 position
                    tokens.insert(-1, '_')
                    token_count = len(tokens)
            else:
                tokens[-2] = '_'

            self.prefix = self.prefix.strip(',')
            self.prefix = Normalize.normalize_for_search(self.prefix, '')
            
        #if token_count > 2:
            #  3rd to last token is Admin2
        #    self.admin2_name, mod = Normalize.admin2_normalize(tokens[-3], self.country_iso)
            
        # fill in country name if still missing - finding Admin1 will find country ISO
        if self.country_name == '' and self.country_iso != '':
            self.country_name = geo_files.geodb.get_country_name(self.country_iso)
            
        # Last two tokens are now Admin1, Country (although they may be '_')
        # If >2 tokens:  Put first non-blank token in City and in Prefix
        # If >3 tokens:  Put second non-blank token in Admin2 and also append to Prefix
        
        # Remove all blank tokens
        tokens = [x for x in tokens if x]
        token_count = len(tokens)

        if token_count > 2:
            # Additional token, take first as city
            self.city1 = Normalize.normalize_for_search(tokens[0], self.country_iso)
            self.target = self.city1
            self.place_type = PlaceType.CITY

            # Also place token[0] into Prefix
            if '*' not in tokens[0]:
                self.prefix += str(tokens[0].strip(' '))

        if token_count > 3:
            # Add ADMIN2
            #  Format: City, Admin2, Admin1, Country
            #  Admin2 is 2nd.  Note -  if Admin2 isnt found, it will look it up as city
            if self.admin2_name == '' or token_count > 4:
                self.admin2_name, modif = Normalize.admin2_normalize(tokens[1], self.country_iso)
            
            #  append token[1] into Prefix
            if '*' not in tokens[1]:
                    self.prefix += ', '
                    self.prefix += str(tokens[1].strip(' '))

        self.logger.debug(f"    ======= PARSE: {place_name} \nCity [{self.city1}] Adm2 [{self.admin2_name}]"
                          f" Adm1 [{self.admin1_name}] adm1_id [{self.admin1_id}] Cntry [{self.country_name}] Pref=[{self.prefix}]"
                          f" type_id={self.place_type}")
        return

    def get_filter_parameters(self, place_name):
        """
            Parse search parameters from place_name and place in appropriate fields.
            Place_name format is:  town, --feature=XXX,--iso=XXX
        # Args:   
            place_name: town, --feature=XXX,--iso=XXX   
        # Returns:   
            Loc fields updated: country_iso, and feature are updated   
        """
        # Separate out arguments
        tokens = place_name.split(",")
        args = []
        for tkn in tokens:
            if '--' in tkn:
                args.append(tkn.strip(' '))

        # Parse options in place name
        parser = ArgumentParserNoExit.ArgumentParserNoExit(description="Parses command.")
        parser.add_argument("-f", "--feature", help=argparse.SUPPRESS)
        parser.add_argument("-i", "--iso", help=argparse.SUPPRESS)
        parser.add_argument("-c", "--country", help=argparse.SUPPRESS)
        try:
            options = parser.parse_args(args)
            self.city1 = Normalize.normalize_for_search(tokens[0], self.country_iso)
            self.target = self.city1
            if options.iso:
                self.country_iso = options.iso.lower()
            if options.country:
                self.country_iso = options.country.lower()
            if options.feature:
                self.feature = options.feature.upper()
            self.place_type = PlaceType.ADVANCED_SEARCH
        except Exception as e:
            self.logger.debug(e)
        self.logger.debug(f'ADV SEARCH: targ={self.city1} iso={self.country_iso} feat={self.feature} typ={self.place_type}')

    def get_status(self) -> str:
        self.logger.debug(f'status=[{self.status}]')
        return self.status

    def set_types_as_string(self):
        # Ensure all items are string type
        self.city1 = str(self.city1)
        self.admin1_name = str(self.admin1_name)
        self.admin2_name = str(self.admin2_name)
        self.prefix = str(self.prefix)

    def add_commas(self, txt) -> str:
        if txt == '':
            if self.need_commas:
                return ', '
            else:
                return ''
        else:
            self.need_commas = True
            return f'{txt}, '

    def get_long_name(self, replace_dct) -> str:
        """
        Take the fields in a Place and build full name.  e.g.  city,adm2,adm1,country name   
        Prefix is NOT included.  Text also has replacements from dictionary applied   
        #Args:   
            replace_dct: Dictionary of text replacements.  'Regex':'replacement'   
        #Returns:  
            long name  
        """
        self.need_commas = False

        if self.admin1_name is None:
            self.admin1_name = ''
        if self.admin2_name is None:
            self.admin2_name = ''

        city = self.add_commas(self.city1)
        admin2 = self.add_commas(self.admin2_name)
        admin1 = self.add_commas(self.admin1_name)

        if self.place_type == PlaceType.COUNTRY:
            nm = f"{self.country_name}"
        elif self.place_type == PlaceType.ADMIN1:
            nm = f"{admin1}{self.country_name}"
        elif self.place_type == PlaceType.ADMIN2:
            nm = f"{admin2}{admin1}{self.country_name}"
        else:
            nm = f"{city}{admin2}{admin1}{str(self.country_name)}"

        # normalize prefix
        self.prefix = Normalize.normalize_for_search(self.prefix, self.country_iso)

        if len(self.prefix) > 0:
            self.prefix_commas = ', '
        else:
            self.prefix_commas = ''

        nm = Loc.capwords(nm)

        # Perform any text replacements user entered into Output Tab
        if replace_dct:
            for key in replace_dct:
                nm = re.sub(key, replace_dct[key], nm)

        return nm

    @staticmethod
    def lowercase_match_group(matchobj):
        return matchobj.group().lower()

    @staticmethod
    def capwords(nm):
        # Change from lowercase to Title Case but fix the title() apostrophe bug
        if nm is not None:
            # Use title(), then fix the title() apostrophe defect
            nm = nm.title()

            # Fix handling for contractions not handled correctly by title()
            poss_regex = r"(?<=[a-z])[\']([A-Z])"
            nm = re.sub(poss_regex, Loc.lowercase_match_group, nm)

        return nm

    def get_five_part_title(self):
        # Returns a five part title string and tokenized version:
        #     prefix,city,county,state,country  (prefix plus long name)

        # Force type to City to generate four part title (then we add prefix for five parts)
        save_type = self.place_type
        self.place_type = PlaceType.CITY

        # Normalize country name
        save_country = self.country_name
        self.country_name, modified = Normalize.country_normalize(self.country_name)

        # if len(self.extra) > 0:
        #    full_title = self.prefix + ' ' + self.extra + ',' + f"{self.city1}, {self.admin2_name}, {self.admin1_name}, {str(self.country_name)}"
        # else:
        full_title = self.prefix + ',' + f"{self.city1}, {self.admin2_name}, {self.admin1_name}, {str(self.country_name)}"

        # Restore values to original
        self.place_type = save_type
        self.country_name = save_country

        return full_title

    def set_type_from_feature(self):
        # Set place type based on DB response feature code
        if self.feature == 'ADM0':
            self.place_type = PlaceType.COUNTRY
        elif self.feature == 'ADM1':
            self.place_type = PlaceType.ADMIN1
        elif self.feature == 'ADM2':
            self.place_type = PlaceType.ADMIN2
        else:
            self.place_type = PlaceType.CITY
        if len(self.prefix) > 0:
            self.place_type = PlaceType.PREFIX
        return self.place_type

    def set_place_type(self):
        # Set place type based on parsing results
        self.place_type = PlaceType.CITY
        if len(str(self.country_name)) > 0:
            self.place_type = PlaceType.COUNTRY
        if len(self.admin1_name) > 0:
            self.place_type = PlaceType.ADMIN1
        if len(self.admin2_name) > 0:
            self.place_type = PlaceType.ADMIN2
        if len(self.city1) > 0:
            self.place_type = PlaceType.CITY

    def set_place_type_text(self):
        # Set result_type_text based on place type
        if self.result_type == GeoUtil.Result.NO_COUNTRY:
            self.result_type_text = 'Country'
        elif self.place_type == PlaceType.COUNTRY:
            self.result_type_text = 'Country'
        elif self.place_type == PlaceType.ADMIN1:
            self.result_type_text = self.get_district1_type(self.country_iso)
        elif self.place_type == PlaceType.ADMIN2:
            self.result_type_text = 'County'
        elif self.place_type == PlaceType.CITY:
            self.result_type_text = self.get_type_name(self.feature)
        elif self.place_type == PlaceType.PREFIX:
            self.result_type_text = 'Place'

    def remove_old_fields(self):
        # Remove fields that are unused by this place type
        if self.place_type == PlaceType.COUNTRY:
            # self.prefix = ''
            self.city1 = ''
            self.admin2_name = ''
            self.admin1_name = ''
        elif self.place_type == PlaceType.ADMIN1:
            # self.prefix = ''
            self.city1 = ''
            self.admin2_name = ''
        elif self.place_type == PlaceType.ADMIN2:
            # self.prefix = ''
            self.city1 = ''
        elif self.place_type == PlaceType.CITY:
            # self.prefix = ''
            pass
        
    @staticmethod
    def sort_words(words: str) -> str:
        word_list:list = words.split(' ')
        word_list.sort()
        return ' '.join(word_list)
    
    @staticmethod
    def get_soundex_by_word(text:str)->str:
        result_sdx = ''
        text_words = text.split(' ')
        for word in text_words:
            if len(result_sdx) == 0:
                result_sdx = GeoUtil.get_soundex(Loc.sort_words(word))
            else:
                result_sdx += ' ' + GeoUtil.get_soundex(Loc.sort_words(word))
        return result_sdx

    @staticmethod
    def matchscore_prefix(pref: str, result: str) -> str:
        """
        Cleanup prefix.  Remove any items from prefix that are in match result.  Remove *   
        #Args:   
            pref:   
            result:   

        #Returns:  Prefix with words removed   
        """
        new_prfx = pref.lower()
        new_prfx = new_prfx.strip(' ')
        prefix_parts = new_prfx.split(',')

        result = result.lower()
        result_parts = result.split(',')

        # Walk thru each comma separated segment in result
        for result_segment_idx, result_segment in enumerate(result_parts):
            result_segment = Loc.sort_words(result_segment)  # Sort words so order doesn't matter
            result_sdx = ' ' + Loc.get_soundex_by_word(result_segment) + ' '
            result_segment = ' ' + result_segment + ' '

            # Walk thru each comma separated segment in prefix
            for prefix_segment_idx, prefix_segment in enumerate(prefix_parts):
                prefix_words = prefix_segment.split(' ')
                
                # Walk thru each word in prefix segment and remove any word that is in result segment
                for pref_word_idx, prefix_word in enumerate(prefix_words):
                    prefix_sdx = ' ' + GeoUtil.get_soundex(prefix_word) + ' '
                    prefix_word_pad = ' ' +  prefix_word + ' '
                    if (prefix_word_pad in result_segment and prefix_word != '') or (prefix_sdx in 
                                                                                 result_sdx and prefix_sdx != ''):
                        # Remove item from result so we don't use again
                        result_segment = re.sub(prefix_word,'', result_segment,1)
                        new_prfx = re.sub(prefix_word,'', new_prfx,1)
                        pass

        res = re.sub('[,]','', new_prfx)
        res = res.strip(' ')

        return res.strip(',')

    @staticmethod
    def prefix_cleanup(pref: str, result: str) -> str:
        """
        Cleanup prefix.  Remove any items from prefix that are in match result.  Remove *   
        #Args:   
            pref:   
            result:   

        #Returns:  Prefix with words removed

        """
        return Loc.matchscore_prefix(pref, result)
        

    @staticmethod
    def get_district1_type(iso) -> str:
        # Return the local country term for Admin1 district
        if iso in ["al", "no"]:
            return "County"
        elif iso in ["us", "at", "bm", "br", "de"]:
            return "State"
        elif iso in ["ac", "an", 'ao', 'bb', 'bd']:
            return "Parish"
        elif iso in ["ae"]:
            return "Emirate"
        elif iso in ["bc", "bf", "bh", "bl", "bn"]:
            return "District"
        elif iso in ["gb"]:
            return "Country"
        else:
            return "Province"

    @staticmethod
    def get_type_name(feature):
        nm = type_name.get(feature)
        if nm is None:
            nm = ''
        return nm


type_name = {
    "ADM0": 'Country', "ADM1": 'City', "ADM2": 'City', "ADM3": 'City', "ADM4": 'City', "ADMF": 'City',
    "CH"  : 'Church', "CSTL": 'Castle', "CMTY": 'Cemetery', "EST": 'Estate', "HSP": 'Hospital',
    "HSTS": 'Historic', "ISL": 'Island', "MSQE": 'Mosque', "MSTY": 'Monastery', "MT": 'Mountain', "MUS": 'Museum', "PAL": 'Palace',
    "PPL" : 'City', "PPLA": 'City', "PPLA2": 'City', "PPLA3": 'City', "PPLA4": 'City',
    "PPLC": 'City', "PPLG": 'City', "PPLH": 'City', "PPLL": 'Village', "PPLQ": 'City', "PPLX": 'City',
    "PRK" : 'Park', "PRN": 'Prison', "PRSH": 'Parish', "RUIN": 'Ruin',
    "RLG" : 'Religious', "STG": '', "SQR": 'Square', "SYG": 'Synagogue', "VAL": 'Valley', "PP1M": 'City', "PP1K": 'City'
    }



#from geodata import Geodata, DB
import importlib
import inspect
import os
import re

def pred(ob):
    return inspect.ismethod(ob) or inspect.isclass(ob) or inspect.isfunction(ob)

def display_modules(cls):
    print(f'-------{cls}---------')
    for item in inspect.getmembers(cls, inspect.isfunction):
        module = item[1]
        if item[0][0] != '_':
            print(f'{item[0]}\n' )
            print(f'{module.__doc__}\n' )
            
base_dir = '/Users/mikeherbert/Documents/'  
            
for package in os.listdir(base_dir):   #os.getcwd()):
    package_path = os.path.join(base_dir, package)
    if os.path.isdir(package_path):
        print(f'=== Package {package} ===')
        for file in os.listdir(package_path):  # os.getcwd()):
            if file.endswith(".py") and file[0] != '_':
                module_nm = re.sub(r'\.py','', file )
                print(f'  {module_nm}.py' )
                    
                module = importlib.import_module(module_nm, package=package)
                
                for item in inspect.getmembers(module, pred):
                    module_name = item[1]
                    if  item[0][0] != '_' and module_nm  in str(inspect.getmodule(item[1])):
                        print(f'      ({item[0]})' )
                        #display_modules(module_name)
                        for itm in inspect.getmembers(item[1], pred):
                            module_name = itm[1]
                            if itm[0][0] != '_' and module_nm  in str(inspect.getmodule(item[1])):
                                print(f'           <{itm[0]}>')
                                #print(f'                {itm[1].__doc__}\n' )
                                pass





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#  Copyright (c) 2019.       Mike Herbert
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
"""    Example program for Geodata gazeteer.  
"""
import logging
import os
import sys
from pathlib import Path

from geodata import Geodata, Loc


class Example:
    """
    Example program for Geodata gazeteer.  
    0. pip3 install geodata
    1. Create folder in home directory:  example/cache  
    2. Download gb.txt from geonames.org and place in example/cache.  (This just has Great Britain data)  
    3. Run example.py  
    """

    def __init__(self):
        # Set up standard logging.  
        logging.getLogger(__name__)
        fmt = "%(levelname)s %(name)s.%(funcName)s %(lineno)d: %(message)s"
        logging.basicConfig(level=logging.DEBUG, stream=sys.stdout, format=fmt)  # Change this to logging.DEBUG for more detail

        # Initialize
        directory = os.path.join(str(Path.home()), "Documents", "geoname_data")
        self.geodata = Geodata.Geodata(directory_name=directory, progress_bar=None, enable_spell_checker=False,
                                       show_message=True, exit_on_error=True,
                                       languages_list_dct={'en'},
                                       feature_code_list_dct=features,
                                       supported_countries_dct={'fr', 'gb', 'ca'})

        # Open Geoname database - city names, lat/long, etc.  Create database if not found
        error = self.geodata.open(repair_database=True, query_limit=105)
        if error:
            print(f"Missing geoname Files in {directory}: download gb.txt or allcountries.txt from geonames.org")
            raise ValueError('Missing files from geonames.org')

    def lookup_place(self, location_name):
        # Create Location instance.  This will hold search parameters and result
        place: Loc.Loc = Loc.Loc()

        # Find best match
        match = self.geodata.find_best_match(location=location_name, place=place)

        if match:
            # Create full name for result
            name = f'{place.get_long_name(None)}'
            print(f'   Best match for {location_name}:\n {name}  Prefix=[{place.prefix}{place.prefix_commas}] Score={place.score:.1f}\n')
        else:
            if place.result_type == Geodata.GeoUtil.Result.NOT_SUPPORTED:
                print(f'   NO match for {location_name}:\n Country {place.country_name} not in supported country list')
            else:
                print(f'   NO match for {location_name}:\n')


# Geoname feature types to add to database.  Other feature types will be ignored.
features = {"ADM1", "ADM2", "ADM3", "ADM4", "ADMF", "CH", "CSTL", "CMTY", "EST ", "HSP", "FT",
            "HSTS", "ISL", "MSQE", "MSTY", "MT", "MUS", "PAL", "PPL", "PPLA", "PPLA2", "PPLA3", "PPLA4",
            "PPLC", "PPLG", "PPLH", "PPLL", "PPLQ", "PPLX", "PRK", "PRN", "PRSH", "RUIN", "RLG", "STG", "SQR", "SYG", "VAL"}

if __name__ == "__main__":
    # Initialize
    ex = Example()

    # Try a few different locations
    locations = [
        '12 baker st, Manchester, , England',  # Street as prefix
        'eddinburg castle,,scotland',  # misspelled
        'cant* cath*,england',  # wildcards
        'd*,--feature=CSTL,--iso=GB',  # search by feature type= castle
        'cardiff, wales',  # good location
        'carddif, wales',  # misspelled 
        'lindering, wales',  # poor match quality
        'phoenix, england',  # doesnt exist
        'Saint-Denis-le-Ferment,,normandie,france',
        'cairo,egypt',
        'tiverton',
        'Thetford Abbey, , england',
        'tretwr, llnfhngl cwm du, breconshire, england,',
        'kathedrale winchester,england',
        ]

    locations2 = [
        'thetford abbey'
        ]

    for name in locations2:
        ex.lookup_place(name)
